@@ -1,68 +1,66 @@
-# -*- coding: utf-8 -*-
-# Licensed under the Apache License, Version 2.0 (the "License");
-# you may not use this file except in compliance with the License.
-# You may obtain a copy of the License at
+# Licensed under the Apache License, Version 2.0 (the "License"); you may
+# not use this file except in compliance with the License. You may obtain
+# a copy of the License at
 #
-#    http://www.apache.org/licenses/LICENSE-2.0
+#   http://www.apache.org/licenses/LICENSE-2.0
 #
 # Unless required by applicable law or agreed to in writing, software
-# distributed under the License is distributed on an "AS IS" BASIS,
-# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
-# implied.
-# See the License for the specific language governing permissions and
-# limitations under the License.
+# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
+# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
+# License for the specific language governing permissions and limitations
+# under the License.
+#
+#  Mistral documentation build configuration file
+#
+# Refer to the Sphinx documentation for advice on configuring this file:
+#
+#   http://www.sphinx-doc.org/en/stable/config.html
 
 import os
 import sys
 
-
 on_rtd = os.environ.get('READTHEDOCS', None) == 'True'
 
 # If extensions (or modules to document with autodoc) are in another directory,
 # add these directories to sys.path here. If the directory is relative to the
 # documentation root, use os.path.abspath to make it absolute, like shown here.
+
 sys.path.insert(0, os.path.abspath('../../'))
 sys.path.insert(0, os.path.abspath('../'))
 sys.path.insert(0, os.path.abspath('./'))
+
 # -- General configuration ----------------------------------------------------
 
 # Add any Sphinx extension module names here, as strings. They can be
 # extensions coming with Sphinx (named 'sphinx.ext.*') or your custom ones.
+
 extensions = [
     'sphinx.ext.autodoc',
-    'sphinxcontrib.autohttp.flask',
     'sphinxcontrib.pecanwsme.rest',
+    'sphinxcontrib.httpdomain',
     'wsmeext.sphinxext',
+    'openstackdocstheme',
+    'oslo_policy.sphinxext',
+    'oslo_policy.sphinxpolicygen',
 ]
 
-if not on_rtd:
-    extensions.append('oslosphinx')
-
 wsme_protocols = ['restjson']
 
-# Add any paths that contain templates here, relative to this directory.
-templates_path = ['_templates']
+suppress_warnings = ['app.add_directive']
 
-# autodoc generation is a bit aggressive and a nuisance when doing heavy
-# text edit cycles.
-# execute "export SPHINX_DEBUG=1" in your terminal to disable
-
-# The suffix of source filenames.
+# The suffix of source file names.
 source_suffix = '.rst'
 
 # The master toctree document.
 master_doc = 'index'
 
 # General information about the project.
-project = u'Mistral'
-copyright = u'2014, Mistral Contributors'
+project = 'Mistral'
+copyright = u'2020, Mistral Contributors'
 
-# The version info for the project you're documenting, acts as replacement for
-# |version| and |release|, also used in various other places throughout the
-# built documents.
-from mistral.version import version_info
-release = version_info.release_string()
-version = version_info.version_string()
+policy_generator_config_file = \
+    '../../tools/config/policy-generator.mistral.conf'
+sample_policy_basename = '_static/mistral'
 
 # If true, sectionauthor and moduleauthor directives will be shown in the
 # output. They are ignored by default.
@@ -76,7 +74,7 @@ add_function_parentheses = True
 add_module_names = True
 
 # The name of the Pygments (syntax highlighting) style to use.
-pygments_style = 'sphinx'
+pygments_style = 'native'
 
 # -- Options for HTML output --------------------------------------------------
 
@@ -84,12 +82,7 @@ pygments_style = 'sphinx'
 # a list of builtin themes.
 # html_static_path = ['_static']
 
-if on_rtd:
-    html_theme_path = ['.']
-    html_theme = 'sphinx_rtd_theme'
-
-# Output file base name for HTML help builder.
-htmlhelp_basename = '%sdoc' % project
+html_theme = 'openstackdocs'
 
 # A list of ignored prefixes for module index sorting.
 modindex_common_prefix = ['mistral.']
@@ -98,17 +91,6 @@ modindex_common_prefix = ['mistral.']
 # "<project> v<release> documentation".
 html_title = 'Mistral'
 
-# Custom sidebar templates, maps document names to template names.
-html_sidebars = {
-    'index': [
-        'sidebarlinks.html', 'localtoc.html', 'searchbox.html', 'sourcelink.html'
-    ],
-    '**': [
-        'localtoc.html', 'relations.html',
-        'searchbox.html', 'sourcelink.html'
-    ]
-}
-
 # -- Options for manual page output -------------------------------------------
 
 # One entry per manual page. List of tuples
@@ -120,3 +102,26 @@ man_pages = [
 
 # If true, show URL addresses after external links.
 man_show_urls = True
+
+# -- Options for openstackdocstheme -------------------------------------------
+
+openstackdocs_repo_name = 'openstack/mistral'
+openstackdocs_pdf_link = True
+openstackdocs_auto_name = False
+openstackdocs_bug_project = 'mistral'
+openstackdocs_bug_tag = 'doc'
+
+latex_use_xindy = False
+
+html_theme_options = {
+    "display_global_toc_section": True,
+    "sidebar_mode": "toctree",
+}
+
+# -- Options for LaTeX output ------------------------------------------------
+
+latex_elements = {
+    'makeindex': '',
+    'printindex': '',
+    'preamble': r'\setcounter{tocdepth}{3}',
+}
