@@ -1,5 +1,3 @@
-# -*- coding: utf-8 -*-
-#
 # Copyright 2013 - Mirantis, Inc.
 #
 #    Licensed under the Apache License, Version 2.0 (the "License");
@@ -21,8 +19,11 @@ from oslo_config import cfg
 from oslo_policy import policy
 
 from mistral import exceptions as exc
+from mistral import policies
 
 
+CONF = cfg.CONF
+
 _ENFORCER = None
 
 
@@ -55,7 +56,7 @@ def enforce(action, context, target=None, do_raise=True,
                    For object creation, this should be a dictionary
                    representing the location of the object.
                    e.g. ``{'project_id': context.project_id}``
-    :param do_raise: if True (the default), raises specified excepion.
+    :param do_raise: if True (the default), raises specified exception.
     :param exc: Exception to be raised if not authorized. Default is
                 mistral.exceptions.NotAllowedException.
 
@@ -63,32 +64,54 @@ def enforce(action, context, target=None, do_raise=True,
              do_raise is False.
     """
 
-    if cfg.CONF.auth_type != 'keystone':
-        # Policy enforcement is supported now only with Keystone
-        # authentication.
-        return
-
     target_obj = {
         'project_id': context.project_id,
         'user_id': context.user_id,
     }
-
     target_obj.update(target or {})
+
+    policy_context = context.to_policy_values()
+    # Because policy.json or policy.yaml example in Mistral repo still uses
+    # the rule 'is_admin: True', we insert 'is_admin' key to the default
+    # policy values.
+    policy_context['is_admin'] = context.is_admin
+
     _ensure_enforcer_initialization()
 
-    return _ENFORCER.enforce(
+    return _ENFORCER.authorize(
         action,
         target_obj,
-        context.to_dict(),
+        policy_context,
         do_raise=do_raise,
         exc=exc
     )
 
 
+def get_enforcer():
+    """Entrypoint that must return the raw oslo.policy enforcer obj.
+
+    This is utilized by the command-line policy tools.
+
+    :returns: :class:`oslo_policy.policy.Enforcer`
+    """
+    # Here we pass an empty list of arguments because there aren't any
+    # arguments that oslo.config or oslo.policy shouldn't already understand
+    # from the CONF object. This makes things easier here because we don't have
+    # to parse arguments passed in from the command line and remove unexpected
+    # arguments before building a Config object.
+    CONF([], project='mistral')
+    enforcer = policy.Enforcer(CONF)
+    enforcer.register_defaults(policies.list_rules())
+    enforcer.load_rules()
+
+    return enforcer
+
+
 def _ensure_enforcer_initialization():
     global _ENFORCER
     if not _ENFORCER:
         _ENFORCER = policy.Enforcer(cfg.CONF)
+        _ENFORCER.register_defaults(policies.list_rules())
         _ENFORCER.load_rules()
 
 
