@@ -1,5 +1,7 @@
 # Copyright 2013 - Mirantis, Inc.
 # Copyright 2018 - Extreme Networks, Inc.
+# Copyright 2019 - NetCracker Technology Corp.
+# Copyright 2020 Nokia Software.
 #
 #    Licensed under the Apache License, Version 2.0 (the "License");
 #    you may not use this file except in compliance with the License.
@@ -19,14 +21,15 @@ from wsme import types as wtypes
 from mistral.api.controllers import resource
 from mistral.api.controllers.v2 import types
 from mistral import exceptions as exc
-from mistral import utils
 from mistral.workflow import states
+from mistral_lib import utils
 
 SCOPE_TYPES = wtypes.Enum(str, 'private', 'public')
 
 
 class ScopedResource(object):
     """Utilities for scoped resources"""
+
     @classmethod
     def validate_scope(cls, scope):
         if scope not in SCOPE_TYPES.values:
@@ -58,7 +61,7 @@ class Workbook(resource.Resource, ScopedResource):
         return cls(id='123e4567-e89b-12d3-a456-426655440000',
                    name='book',
                    definition='HERE GOES'
-                        'WORKBOOK DEFINITION IN MISTRAL DSL v2',
+                              'WORKBOOK DEFINITION IN MISTRAL DSL v2',
                    tags=['large', 'expensive'],
                    scope='private',
                    project_id='a7eb669e9819420ea4bd1453e672c0a7',
@@ -89,9 +92,10 @@ class Workflow(resource.Resource, ScopedResource):
     name = wtypes.text
     namespace = wtypes.text
     input = wtypes.text
-
+    interface = types.jsontype
+    "input and output of the workflow"
     definition = wtypes.text
-    "Workflow definition in Mistral v2 DSL"
+    "workflow text written in Mistral v2 language"
     tags = [wtypes.text]
     scope = SCOPE_TYPES
     "'private' or 'public'"
@@ -106,16 +110,23 @@ class Workflow(resource.Resource, ScopedResource):
                    name='flow',
                    input='param1, param2',
                    definition='HERE GOES'
-                        'WORKFLOW DEFINITION IN MISTRAL DSL v2',
+                              'WORKFLOW DEFINITION IN MISTRAL DSL v2',
                    tags=['large', 'expensive'],
                    scope='private',
                    project_id='a7eb669e9819420ea4bd1453e672c0a7',
                    created_at='1970-01-01T00:00:00.000000',
                    updated_at='1970-01-01T00:00:00.000000',
-                   namespace='')
+                   namespace='',
+                   interface={"input": ["param1", {"param2": 2}],
+                              "output": []}
+                   )
 
-    @classmethod
-    def _set_input(cls, obj, wf_spec):
+    def set_attributes_from_spec(self, wf_spec):
+        #  Sets input and interface fields for the Workflow resource.
+        self._set_input(wf_spec)
+        self._set_interface(wf_spec)
+
+    def _set_input(self, wf_spec):
         input_list = []
 
         if wf_spec:
@@ -124,39 +135,52 @@ class Workflow(resource.Resource, ScopedResource):
             for param in input:
                 if isinstance(param, dict):
                     for k, v in param.items():
-                        input_list.append("%s=%s" % (k, v))
+                        input_list.append('%s="%s"' % (k, v))
                 else:
                     input_list.append(param)
 
-            setattr(obj, 'input', ", ".join(input_list) if input_list else '')
+            self.input = ", ".join(input_list) if input_list else ''
 
-        return obj
+    def _set_interface(self, wf_spec):
+        self.interface = {}
+
+        if wf_spec:
+            self.interface['input'] = wf_spec.get('input', [])
+            self.interface['output'] = [
+                output for output in wf_spec.get('output', {})
+            ]
 
     @classmethod
     def from_dict(cls, d):
         obj = super(Workflow, cls).from_dict(d)
 
-        return cls._set_input(obj, d.get('spec'))
+        obj.set_attributes_from_spec(d.get('spec'))
+
+        return obj
 
     @classmethod
     def from_db_model(cls, db_model):
         obj = super(Workflow, cls).from_db_model(db_model)
 
-        return cls._set_input(obj, db_model.spec)
+        obj.set_attributes_from_spec(db_model.get('spec'))
+
+        return obj
 
     @classmethod
     def from_tuples(cls, tuple_iterator):
         obj = cls()
         spec = None
+
         for col_name, col_val in tuple_iterator:
             if hasattr(obj, col_name):
                 # Convert all datetime values to strings.
                 setattr(obj, col_name, utils.datetime_to_str(col_val))
+
             if col_name == 'spec':
                 spec = col_val
 
         if spec:
-            obj = cls._set_input(obj, spec)
+            obj.set_attributes_from_spec(spec)
 
         return obj
 
@@ -205,6 +229,7 @@ class Action(resource.Resource, ScopedResource):
 
     created_at = wtypes.text
     updated_at = wtypes.text
+    namespace = wtypes.text
 
     @classmethod
     def sample(cls):
@@ -216,7 +241,8 @@ class Action(resource.Resource, ScopedResource):
             scope='private',
             project_id='a7eb669e9819420ea4bd1453e672c0a7',
             created_at='1970-01-01T00:00:00.000000',
-            updated_at='1970-01-01T00:00:00.000000'
+            updated_at='1970-01-01T00:00:00.000000',
+            namespace=''
         )
 
 
@@ -271,6 +297,9 @@ class Execution(resource.Resource):
     description = wtypes.text
     "description of workflow execution"
 
+    tags = [wtypes.text]
+    "tags of workflow execution"
+
     params = types.jsontype
     """'params' define workflow type specific parameters. Specific parameters
      are:
@@ -314,6 +343,8 @@ class Execution(resource.Resource):
 
     project_id = wsme.wsattr(wtypes.text, readonly=True)
 
+    published_global = types.jsontype
+
     @classmethod
     def sample(cls):
         return cls(
@@ -322,10 +353,12 @@ class Execution(resource.Resource):
             workflow_namespace='some_namespace',
             workflow_id='123e4567-e89b-12d3-a456-426655441111',
             description='this is the first execution.',
+            tags=['simple', 'amazing'],
             project_id='40a908dbddfe48ad80a87fb30fa70a03',
             state='SUCCESS',
             input={},
             output={},
+            published_global={'key': 'value'},
             params={
                 'env': {'k1': 'abc', 'k2': 123},
                 'notify': [
@@ -386,6 +419,8 @@ class Task(resource.Resource):
     workflow_id = wtypes.text
     workflow_execution_id = wtypes.text
 
+    tags = [wtypes.text]
+
     state = wtypes.text
     """state can take one of the following values:
      IDLE, RUNNING, SUCCESS, ERROR, DELAYED"""
@@ -399,11 +434,15 @@ class Task(resource.Resource):
 
     result = wtypes.text
     published = types.jsontype
+    published_global = types.jsontype
     processed = bool
 
     created_at = wtypes.text
     updated_at = wtypes.text
 
+    started_at = wtypes.text
+    finished_at = wtypes.text
+
     # Add this param to make Mistral API work with WSME 0.8.0 or higher version
     reset = wsme.wsattr(bool, mandatory=True)
 
@@ -416,6 +455,7 @@ class Task(resource.Resource):
             workflow_name='flow',
             workflow_id='123e4567-e89b-12d3-a456-426655441111',
             workflow_execution_id='123e4567-e89b-12d3-a456-426655440000',
+            tags=['long', 'security'],
             name='task',
             state=states.SUCCESS,
             project_id='40a908dbddfe48ad80a87fb30fa70a03',
@@ -429,6 +469,7 @@ class Task(resource.Resource):
             },
             result='task result',
             published={'key': 'value'},
+            published_global={'key': 'value'},
             processed=True,
             created_at='1970-01-01T00:00:00.000000',
             updated_at='1970-01-01T00:00:00.000000',
@@ -718,3 +759,163 @@ class EventTriggers(resource.ResourceList):
                                 "marker=123e4567-e89b-12d3-a456-426655440000")
 
         return triggers_sample
+
+
+class BaseExecutionReportEntry(resource.Resource):
+    """Execution report entry resource."""
+
+    id = wtypes.text
+    name = wtypes.text
+    created_at = wtypes.text
+    updated_at = wtypes.text
+    state = wtypes.text
+    state_info = wtypes.text
+
+    @classmethod
+    def sample(cls):
+        # TODO(rakhmerov): complete
+
+        return cls(
+            id='123e4567-e89b-12d3-a456-426655441414',
+            created_at='2019-01-30T00:00:00.000000',
+            updated_at='2019-01-30T00:00:00.000000',
+            state=states.SUCCESS
+        )
+
+
+class ActionExecutionReportEntry(BaseExecutionReportEntry):
+    """Action execution report entry resource."""
+
+    accepted = bool
+    last_heartbeat = wtypes.text
+
+    @classmethod
+    def sample(cls):
+        sample = super(ActionExecutionReportEntry, cls).sample()
+
+        sample.accepted = True
+        sample.last_heartbeat = '2019-01-30T00:00:00.000000'
+
+        return sample
+
+
+class WorkflowExecutionReportEntry(BaseExecutionReportEntry):
+    """Workflow execution report entry resource."""
+
+    # NOTE(rakhmerov): task_executions has to be declared below
+    # after we declare a class for task execution entry resource.
+
+    @classmethod
+    def sample(cls):
+        sample = super(WorkflowExecutionReportEntry, cls).sample()
+
+        # We can't define a non-empty list task executions here because
+        # the needed class is not defined yet. Since this is just a sample
+        # we can sacrifice it.
+        sample.task_executions = []
+
+        return sample
+
+
+class TaskExecutionReportEntry(BaseExecutionReportEntry):
+    """Task execution report entity resource."""
+
+    action_executions = [ActionExecutionReportEntry]
+    workflow_executions = [WorkflowExecutionReportEntry]
+    retry_count = wtypes.IntegerType(minimum=0)
+
+    @classmethod
+    def sample(cls):
+        sample = super(TaskExecutionReportEntry, cls).sample()
+
+        sample.action_executions = [ActionExecutionReportEntry.sample()]
+        sample.workflow_executions = []
+        sample.retry_count = 0
+
+        return sample
+
+
+# We have to declare this field later because of the dynamic binding.
+# It can't be within WorkflowExecutionReportEntry before
+# TaskExecutionReportEntry is declared.
+WorkflowExecutionReportEntry.task_executions = [TaskExecutionReportEntry]
+wtypes.registry.reregister(WorkflowExecutionReportEntry)
+
+
+class ExecutionReportStatistics(resource.Resource):
+    """Execution report statistics.
+
+    TODO(rakhmerov): There's much more we can add here. For example,
+    information about action, average (and also min and max) task execution
+    run time etc.
+    """
+
+    total_tasks_count = wtypes.IntegerType(minimum=0)
+    running_tasks_count = wtypes.IntegerType(minimum=0)
+    success_tasks_count = wtypes.IntegerType(minimum=0)
+    error_tasks_count = wtypes.IntegerType(minimum=0)
+    idle_tasks_count = wtypes.IntegerType(minimum=0)
+    paused_tasks_count = wtypes.IntegerType(minimum=0)
+    estimated_time = wtypes.IntegerType(minimum=-1)
+
+    def __init__(self, **kw):
+        self.total_tasks_count = 0
+        self.running_tasks_count = 0
+        self.success_tasks_count = 0
+        self.error_tasks_count = 0
+        self.idle_tasks_count = 0
+        self.paused_tasks_count = 0
+        self.estimated_time = 0
+
+        super(ExecutionReportStatistics, self).__init__(**kw)
+
+    def increment_running(self):
+        self.running_tasks_count += 1
+        self.total_tasks_count += 1
+
+    def increment_success(self):
+        self.success_tasks_count += 1
+        self.total_tasks_count += 1
+
+    def increment_error(self):
+        self.error_tasks_count += 1
+        self.total_tasks_count += 1
+
+    def increment_idle(self):
+        self.idle_tasks_count += 1
+        self.total_tasks_count += 1
+
+    def increment_paused(self):
+        self.paused_tasks_count += 1
+        self.total_tasks_count += 1
+
+    @classmethod
+    def sample(cls):
+        return cls(
+            total_tasks_count=10,
+            running_tasks_count=3,
+            success_tasks_count=5,
+            error_tasks_count=2,
+            idle_tasks_count=0,
+            paused_tasks_count=0,
+            estimated_time=40
+        )
+
+
+class ExecutionReport(resource.Resource):
+    """Execution report resource."""
+
+    statistics = ExecutionReportStatistics
+    """General statistics about the workflow execution hierarchy."""
+
+    root_workflow_execution = WorkflowExecutionReportEntry
+    """Root entry of the report associated with a workflow execution."""
+
+    @classmethod
+    def sample(cls):
+        sample = cls()
+
+        sample.statistics = ExecutionReportStatistics.sample()
+        sample.root_workflow_execution = WorkflowExecutionReportEntry.sample()
+
+        return sample
