@@ -1,5 +1,6 @@
 # Copyright 2013 - Mirantis, Inc.
 # Copyright 2015 - StackStorm, Inc.
+# Copyright 2019 - NetCracker Technology Corp.
 #
 #    Licensed under the Apache License, Version 2.0 (the "License");
 #    you may not use this file except in compliance with the License.
@@ -23,10 +24,12 @@ import wsmeext.pecan as wsme_pecan
 from mistral.api import access_control as acl
 from mistral.api.controllers.v2 import action_execution
 from mistral.api.controllers.v2 import resources
+from mistral.api.controllers.v2 import sub_execution
 from mistral.api.controllers.v2 import types
 from mistral import context
 from mistral.db.v2 import api as db_api
 from mistral import exceptions as exc
+from mistral import expressions as expr
 from mistral.lang import parser as spec_parser
 from mistral.rpc import clients as rpc
 from mistral.utils import filter_utils
@@ -34,10 +37,16 @@ from mistral.utils import rest_utils
 from mistral.workflow import data_flow
 from mistral.workflow import states
 
-
 LOG = logging.getLogger(__name__)
-STATE_TYPES = wtypes.Enum(str, states.IDLE, states.RUNNING, states.SUCCESS,
-                          states.ERROR, states.RUNNING_DELAYED)
+
+STATE_TYPES = wtypes.Enum(
+    str,
+    states.IDLE,
+    states.RUNNING,
+    states.SUCCESS,
+    states.ERROR,
+    states.RUNNING_DELAYED
+)
 
 
 def _get_task_resource_with_result(task_ex):
@@ -48,18 +57,75 @@ def _get_task_resource_with_result(task_ex):
     return task
 
 
+# Use retries to prevent possible failures.
+@rest_utils.rest_retry_on_db_error
+def _get_task_execution(id):
+    with db_api.transaction():
+        task_ex = db_api.get_task_execution(id)
+
+        rest_utils.load_deferred_fields(task_ex, ['workflow_execution'])
+        rest_utils.load_deferred_fields(
+            task_ex.workflow_execution,
+            ['context', 'input', 'params', 'root_execution']
+        )
+
+        rest_utils.load_deferred_fields(
+            task_ex.workflow_execution.root_execution,
+            ['params']
+        )
+
+        return _get_task_resource_with_result(task_ex), task_ex
+
+
+def get_published_global(task_ex, wf_ex=None):
+    if task_ex.state not in [states.SUCCESS, states.ERROR]:
+        return
+
+    if wf_ex is None:
+        wf_ex = task_ex.workflow_execution
+
+    expr_ctx = data_flow.ContextView(
+        data_flow.get_current_task_dict(task_ex),
+        task_ex.in_context,
+        data_flow.get_workflow_environment_dict(wf_ex),
+        wf_ex.context,
+        wf_ex.input
+    )
+
+    task_spec = spec_parser.get_task_spec(task_ex.spec)
+    publish_spec = task_spec.get_publish(task_ex.state)
+
+    if not publish_spec:
+        return
+
+    global_vars = publish_spec.get_global()
+
+    return expr.evaluate_recursively(global_vars, expr_ctx)
+
+
+def _task_with_published_global(task, task_ex):
+    published_global_vars = get_published_global(task_ex)
+
+    if published_global_vars:
+        task.published_global = published_global_vars
+
+    return task
+
+
 class TaskExecutionsController(rest.RestController):
     @rest_utils.wrap_wsme_controller_exception
     @wsme_pecan.wsexpose(resources.Executions, types.uuid, types.uuid, int,
                          types.uniquelist, types.list, types.uniquelist,
-                         wtypes.text, types.uuid, wtypes.text, types.jsontype,
-                         STATE_TYPES, wtypes.text, types.jsontype,
-                         types.jsontype, wtypes.text, wtypes.text)
+                         wtypes.text, types.uuid, wtypes.text,
+                         types.uniquelist, types.jsontype, STATE_TYPES,
+                         wtypes.text, types.jsontype, types.jsontype,
+                         wtypes.text, wtypes.text)
     def get_all(self, task_execution_id, marker=None, limit=None,
                 sort_keys='created_at', sort_dirs='asc', fields='',
                 workflow_name=None, workflow_id=None, description=None,
-                params=None, state=None, state_info=None, input=None,
-                output=None, created_at=None, updated_at=None):
+                tags=None, params=None, state=None,
+                state_info=None, input=None, output=None,
+                created_at=None, updated_at=None):
         """Return all executions that belong to the given task execution.
 
         :param task_execution_id: Task task execution ID.
@@ -83,6 +149,7 @@ class TaskExecutionsController(rest.RestController):
                             workflow ID.
         :param description: Optional. Keep only resources with a specific
                             description.
+        :param tags: Optional. Keep only resources containing specific tags.
         :param params: Optional. Keep only resources with specific parameters.
         :param state: Optional. Keep only resources with a specific state.
         :param state_info: Optional. Keep only resources with specific
@@ -101,6 +168,7 @@ class TaskExecutionsController(rest.RestController):
             created_at=created_at,
             workflow_name=workflow_name,
             workflow_id=workflow_id,
+            tags=tags,
             params=params,
             state=state,
             state_info=state_info,
@@ -130,18 +198,10 @@ class TaskExecutionsController(rest.RestController):
         )
 
 
-# Use retries to prevent possible failures.
-@rest_utils.rest_retry_on_db_error
-def _get_task_execution(id):
-    with db_api.transaction():
-        task_ex = db_api.get_task_execution(id)
-
-        return _get_task_resource_with_result(task_ex)
-
-
 class TasksController(rest.RestController):
     action_executions = action_execution.TasksActionExecutionController()
     workflow_executions = TaskExecutionsController()
+    executions = sub_execution.SubExecutionsController()
 
     @rest_utils.wrap_wsme_controller_exception
     @wsme_pecan.wsexpose(resources.Task, wtypes.text)
@@ -153,19 +213,25 @@ class TasksController(rest.RestController):
         acl.enforce('tasks:get', context.ctx())
         LOG.debug("Fetch task [id=%s]", id)
 
-        return _get_task_execution(id)
+        task, task_ex = _get_task_execution(id)
+
+        return _task_with_published_global(task, task_ex)
 
     @rest_utils.wrap_wsme_controller_exception
     @wsme_pecan.wsexpose(resources.Tasks, types.uuid, int, types.uniquelist,
                          types.list, types.uniquelist, wtypes.text,
-                         wtypes.text, types.uuid, types.uuid, STATE_TYPES,
-                         wtypes.text, wtypes.text, types.jsontype, bool,
-                         wtypes.text, wtypes.text, bool, types.jsontype)
+                         wtypes.text, types.uuid,
+                         types.uuid, types.uniquelist, STATE_TYPES,
+                         wtypes.text, wtypes.text, types.jsontype,
+                         bool, wtypes.text, wtypes.text,
+                         bool, types.jsontype)
     def get_all(self, marker=None, limit=None, sort_keys='created_at',
-                sort_dirs='asc', fields='', name=None, workflow_name=None,
-                workflow_id=None, workflow_execution_id=None, state=None,
-                state_info=None, result=None, published=None, processed=None,
-                created_at=None, updated_at=None, reset=None, env=None):
+                sort_dirs='asc', fields='', name=None,
+                workflow_name=None, workflow_id=None,
+                workflow_execution_id=None, tags=None, state=None,
+                state_info=None, result=None, published=None,
+                processed=None, created_at=None, updated_at=None,
+                reset=None, env=None):
         """Return all tasks.
 
         Where project_id is the same as the requester or
@@ -214,6 +280,7 @@ class TasksController(rest.RestController):
             created_at=created_at,
             workflow_name=workflow_name,
             workflow_id=workflow_id,
+            tags=tags,
             state=state,
             state_info=state_info,
             updated_at=updated_at,
@@ -318,13 +385,15 @@ class ExecutionTasksController(rest.RestController):
     @rest_utils.wrap_wsme_controller_exception
     @wsme_pecan.wsexpose(resources.Tasks, types.uuid, types.uuid, int,
                          types.uniquelist, types.list, types.uniquelist,
-                         wtypes.text, wtypes.text, types.uuid, STATE_TYPES,
-                         wtypes.text, wtypes.text, types.jsontype, bool,
+                         wtypes.text, wtypes.text, types.uuid,
+                         types.uniquelist, STATE_TYPES, wtypes.text,
+                         wtypes.text, types.jsontype, bool,
                          wtypes.text, wtypes.text, bool, types.jsontype)
     def get_all(self, workflow_execution_id, marker=None, limit=None,
-                sort_keys='created_at', sort_dirs='asc', fields='', name=None,
-                workflow_name=None, workflow_id=None, state=None,
-                state_info=None, result=None, published=None, processed=None,
+                sort_keys='created_at', sort_dirs='asc', fields='',
+                name=None, workflow_name=None, workflow_id=None,
+                tags=None, state=None, state_info=None,
+                result=None, published=None, processed=None,
                 created_at=None, updated_at=None, reset=None, env=None):
         """Return all tasks within the execution.
 
@@ -352,6 +421,7 @@ class ExecutionTasksController(rest.RestController):
                             workflow ID.
         :param workflow_execution_id: Optional. Keep only resources with a
                                       specific workflow execution ID.
+        :param tags: Optional. Keep only resources containing specific tags.
         :param state: Optional. Keep only resources with a specific state.
         :param state_info: Optional. Keep only resources with specific
                            state information.
@@ -375,6 +445,7 @@ class ExecutionTasksController(rest.RestController):
             created_at=created_at,
             workflow_name=workflow_name,
             workflow_id=workflow_id,
+            tags=tags,
             state=state,
             state_info=state_info,
             updated_at=updated_at,
