@@ -21,7 +21,6 @@ Configuration options registration and useful routines.
 
 import itertools
 import json
-import os
 
 from keystoneauth1 import loading
 from oslo_config import cfg
@@ -32,6 +31,7 @@ from osprofiler import opts as profiler
 from mistral import version
 
 from mistral._i18n import _
+from mistral.workflow import states
 
 # Options under default group.
 launch_opt = cfg.ListOpt(
@@ -63,6 +63,35 @@ auth_type_opt = cfg.StrOpt(
     help=_('Authentication type (valid options: keystone, keycloak-oidc)')
 )
 
+legacy_action_provider_opts = [
+    cfg.BoolOpt(
+        'load_action_plugins',
+        default=True,
+        help=_(
+            'If True, enables loading actions configured in the '
+            'entry point "mistral.actions".'
+        )
+    ),
+    cfg.BoolOpt(
+        'load_action_generators',
+        default=True,
+        help=_(
+            'If True, enables loading actions from action generators '
+            'configured in the entry point "mistral.generators".'
+        )
+    ),
+    cfg.BoolOpt(
+        'only_builtin_actions',
+        default=False,
+        help=_(
+            'If True, then the legacy action provider loads only '
+            'the actions delivered by the Mistral project out of '
+            'the box plugged in with the entry point "mistral.actions".'
+            'This property is needed mostly for testing.'
+        )
+    ),
+]
+
 api_opts = [
     cfg.HostAddressOpt(
         'host',
@@ -318,7 +347,7 @@ scheduler_opts = [
              'locally before it becomes eligible for processing by other '
              'scheduler instances.'
              'For example, a job needs to run at 12:00:00. When a scheduler '
-             'starts processing it it has 60 seconds (or other configured '
+             'starts processing it has 60 seconds (or other configured '
              'value) to complete the job. If the scheduler did not complete '
              'the job within this period it most likely means that the '
              'scheduler process crashed. In this case another scheduler '
@@ -342,12 +371,12 @@ cron_trigger_opts = [
         default=1,
         min=1,
         help=(
-            'This setting defines how frequently Mistral checks for cron ',
-            'triggers that need execution. By default this is every second ',
-            'which can lead to high system load. Increasing the number will ',
-            'reduce the load but also limit the minimum freqency. For ',
-            'example, a cron trigger can be configured to run every second ',
-            'but if the execution_interval is set to 60, it will only run ',
+            'This setting defines how frequently Mistral checks for cron '
+            'triggers that need execution. By default this is every second '
+            'which can lead to high system load. Increasing the number will '
+            'reduce the load but also limit the minimum freqency. For '
+            'example, a cron trigger can be configured to run every second '
+            'but if the execution_interval is set to 60, it will only run '
             'once per minute.'
         )
     )
@@ -445,6 +474,13 @@ execution_expiration_policy_opts = [
                'The default value is 0. If it is set to 0, '
                'size of batch is total number of expired executions '
                'that is going to be deleted.')
+    ),
+    cfg.ListOpt(
+        'ignored_states',
+        default=[],
+        help='The states that the expiration policy will filter '
+             'out and will not delete.'
+             'Valid values are, [{}]'.format(states.TERMINAL_STATES)
     )
 ]
 
@@ -538,41 +574,18 @@ keycloak_oidc_opts = [
         default='/realms/%s/protocol/openid-connect/userinfo',
         help='Endpoint against which authorization will be performed'
     ),
-]
-
-openstack_actions_opts = [
     cfg.StrOpt(
-        'os-actions-endpoint-type',
-        default=os.environ.get('OS_ACTIONS_ENDPOINT_TYPE', 'public'),
-        choices=['public', 'admin', 'internal'],
-        deprecated_group='DEFAULT',
-        help=_('Type of endpoint in identity service catalog to use for'
-               ' communication with OpenStack services.')
-    ),
-    cfg.ListOpt(
-        'modules-support-region',
-        default=['nova', 'glance', 'heat', 'neutron', 'cinder',
-                 'trove', 'ironic', 'designate', 'murano', 'tacker', 'senlin',
-                 'aodh', 'gnocchi'],
-        help=_('List of module names that support region in actions.')
+        'public_cert_url',
+        default="/realms/%s/protocol/openid-connect/certs",
+        help="URL to get the public key for a particular realm"
     ),
     cfg.StrOpt(
-        'default_region',
-        help=_('Default region name for openstack actions supporting region.')
-    ),
+        'keycloak_iss',
+        help="Keycloak issuer(iss) url. "
+             "Example: https://ip_add:port/auth/realms/%s"
+    )
 ]
 
-# note: this command line option is used only from sync_db and
-# mistral-db-manage
-os_actions_mapping_path = cfg.StrOpt(
-    'openstack_actions_mapping_path',
-    short='m',
-    metavar='MAPPING_PATH',
-    default='actions/openstack/mapping.json',
-    help='Path to openstack action mapping json file.'
-         'It could be relative to mistral package '
-         'directory or absolute.'
-)
 
 yaql_opts = [
     cfg.IntOpt(
@@ -597,7 +610,7 @@ yaql_opts = [
         'convert_input_data',
         default=True,
         help=_('Enables input data conversion for YAQL expressions. If set '
-               'to True then YAQL will convert mutable data structures '
+               'to True, YAQL will convert mutable data structures '
                '(lists, dicts, sets) into their immutable versions. That '
                'will allow them to work with some constructs that require '
                'hashable types even if elements are not hashable. For '
@@ -610,23 +623,38 @@ yaql_opts = [
                'performance boost if the input data for an expression is '
                'large.')
     ),
+    cfg.BoolOpt(
+        'convert_output_data',
+        default=True,
+        help=_('Enables output data conversion for YAQL expressions.'
+               'If set to False, it is possible that YAQL will generate '
+               'an output that will be not JSON-serializable. For example, '
+               'if an expression has ".toSet()" in the end to convert a list '
+               'into a set. It does not mean though that such functions '
+               'cannot be used, they can still be used in expressions but '
+               'user has to keep in mind of what type a result will be, '
+               'whereas if the value of ths property is True YAQL will '
+               'convert the result to a JSON-compatible type.')
+    ),
     cfg.BoolOpt(
         'convert_tuples_to_lists',
         default=True,
-        help=_('When set to true, yaql converts all tuples in the expression '
-               'result to lists.')
+        help=_('When set to True, yaql converts all tuples in the expression '
+               'result to lists. It works only if "convert_output_data" is '
+               'set to True.')
     ),
     cfg.BoolOpt(
         'convert_sets_to_lists',
         default=False,
-        help=_('When set to true, yaql converts all sets in the expression '
+        help=_('When set to True, yaql converts all sets in the expression '
                'result to lists. Otherwise the produced result may contain '
-               'sets that are not JSON-serializable.')
+               'sets that are not JSON-serializable. It works only if '
+               '"convert_output_data" is set to True.')
     ),
     cfg.BoolOpt(
         'iterable_dicts',
         default=False,
-        help=_('When set to true, dictionaries are considered to be iterable '
+        help=_('When set to True, dictionaries are considered to be iterable '
                'and iteration over dictionaries produces their keys (as in '
                'Python and yaql 0.2).')
     ),
@@ -646,6 +674,7 @@ yaql_opts = [
 
 CONF = cfg.CONF
 
+LEGACY_ACTION_PROVIDER_GROUP = 'legacy_action_provider'
 API_GROUP = 'api'
 ENGINE_GROUP = 'engine'
 EXECUTOR_GROUP = 'executor'
@@ -659,7 +688,6 @@ EXECUTION_EXPIRATION_POLICY_GROUP = 'execution_expiration_policy'
 ACTION_HEARTBEAT_GROUP = 'action_heartbeat'
 PROFILER_GROUP = profiler.list_opts()[0][0]
 KEYCLOAK_OIDC_GROUP = "keycloak_oidc"
-OPENSTACK_ACTIONS_GROUP = 'openstack_actions'
 YAQL_GROUP = "yaql"
 KEYSTONE_GROUP = "keystone"
 
@@ -673,6 +701,10 @@ CONF.register_opt(rpc_response_timeout_opt)
 CONF.register_opt(oslo_rpc_executor)
 CONF.register_opt(expiration_token_duration)
 
+CONF.register_opts(
+    legacy_action_provider_opts,
+    group=LEGACY_ACTION_PROVIDER_GROUP
+)
 CONF.register_opts(api_opts, group=API_GROUP)
 CONF.register_opts(engine_opts, group=ENGINE_GROUP)
 CONF.register_opts(executor_opts, group=EXECUTOR_GROUP)
@@ -692,7 +724,6 @@ CONF.register_opts(pecan_opts, group=PECAN_GROUP)
 CONF.register_opts(coordination_opts, group=COORDINATION_GROUP)
 CONF.register_opts(profiler_opts, group=PROFILER_GROUP)
 CONF.register_opts(keycloak_oidc_opts, group=KEYCLOAK_OIDC_GROUP)
-CONF.register_opts(openstack_actions_opts, group=OPENSTACK_ACTIONS_GROUP)
 CONF.register_opts(yaql_opts, group=YAQL_GROUP)
 loading.register_session_conf_options(CONF, KEYSTONE_GROUP)
 
@@ -715,8 +746,6 @@ default_group_opts = itertools.chain(
     ]
 )
 
-CONF.register_cli_opts(CLI_OPTS)
-
 
 _DEFAULT_LOG_LEVELS = [
     'eventlet.wsgi.server=WARN',
@@ -742,7 +771,6 @@ def list_opts():
         (EXECUTION_EXPIRATION_POLICY_GROUP, execution_expiration_policy_opts),
         (PROFILER_GROUP, profiler_opts),
         (KEYCLOAK_OIDC_GROUP, keycloak_oidc_opts),
-        (OPENSTACK_ACTIONS_GROUP, openstack_actions_opts),
         (YAQL_GROUP, yaql_opts),
         (ACTION_HEARTBEAT_GROUP, action_heartbeat_opts),
         (None, default_group_opts)
