@@ -1,6 +1,7 @@
 # Copyright 2015 - Mirantis, Inc.
 # Copyright 2015 - StackStorm, Inc.
 # Copyright 2016 - Brocade Communications Systems, Inc.
+# Copyright 2020 Nokia Software.
 #
 #    Licensed under the Apache License, Version 2.0 (the "License");
 #    you may not use this file except in compliance with the License.
@@ -15,6 +16,8 @@
 #    limitations under the License.
 
 import contextlib
+import datetime
+import re
 import sys
 import threading
 
@@ -35,14 +38,12 @@ from mistral.db.v2.sqlalchemy import filters as db_filters
 from mistral.db.v2.sqlalchemy import models
 from mistral import exceptions as exc
 from mistral.services import security
-from mistral import utils
 from mistral.workflow import states
-
+from mistral_lib import utils
 
 CONF = cfg.CONF
 LOG = logging.getLogger(__name__)
 
-
 _SCHEMA_LOCK = threading.RLock()
 _initialized = False
 
@@ -123,6 +124,11 @@ def refresh(model, session=None):
     session.refresh(model)
 
 
+@b.session_aware()
+def expire_all(session=None):
+    session.expire_all()
+
+
 @b.session_aware()
 def acquire_lock(model, id, session=None):
     # Expire all so all objects queried after lock is acquired
@@ -251,7 +257,9 @@ def _delete_all(model, **kwargs):
     # NOTE(kong): Because we use 'in_' operator in _secure_query(), delete()
     # method will raise error with default parameter. Please refer to
     # http://docs.sqlalchemy.org/en/rel_1_0/orm/query.html#sqlalchemy.orm.query.Query.delete
-    _secure_query(model).filter_by(**kwargs).delete(synchronize_session=False)
+    query = _secure_query(model)
+    query = db_filters.apply_filters(query, model, **kwargs)
+    query.delete(synchronize_session=False)
 
 
 def _get_collection(model, insecure=False, limit=None, marker=None,
@@ -277,6 +285,14 @@ def _get_collection(model, insecure=False, limit=None, marker=None,
     return query.all()
 
 
+def _get_count(model, insecure=False, **filters):
+    query = b.model_query(model) if insecure else _secure_query(model)
+
+    query = db_filters.apply_filters(query, model, **filters)
+
+    return query.count()
+
+
 def _get_db_object_by_name(model, name, columns=()):
     query = _secure_query(model, *columns)
 
@@ -320,7 +336,7 @@ def _get_db_object_by_name_and_namespace_or_id(model, identifier,
 
 
 def _get_db_object_by_name_and_namespace(model, name,
-                                         namespace, insecure=False,
+                                         namespace='', insecure=False,
                                          columns=()):
     query = (
         b.model_query(model, columns=columns)
@@ -636,26 +652,38 @@ def get_action_definition_by_id(id, fields=(), session=None):
 
 
 @b.session_aware()
-def get_action_definition(identifier, fields=(), session=None):
+def get_action_definition(identifier, fields=(), session=None, namespace=''):
     a_def = _get_db_object_by_name_and_namespace_or_id(
         models.ActionDefinition,
         identifier,
+        namespace=namespace,
         columns=fields
     )
+    #  If the action was not found in the given namespace,
+    #  look in the default namespace
+    if not a_def:
+        a_def = _get_db_object_by_name_and_namespace_or_id(
+            models.ActionDefinition,
+            identifier,
+            namespace='',
+            columns=fields
+        )
 
     if not a_def:
         raise exc.DBEntityNotFoundError(
-            "Action definition not found [action_name=%s]" % identifier
+            "Action definition not found [action_name=%s,namespace=%s]"
+            % (identifier, namespace)
         )
 
     return a_def
 
 
 @b.session_aware()
-def load_action_definition(name, fields=(), session=None):
-    return _get_db_object_by_name(
+def load_action_definition(name, fields=(), session=None, namespace=''):
+    return _get_db_object_by_name_and_namespace(
         models.ActionDefinition,
         name,
+        namespace=namespace,
         columns=fields
     )
 
@@ -675,8 +703,8 @@ def create_action_definition(values, session=None):
         a_def.save(session=session)
     except db_exc.DBDuplicateEntry:
         raise exc.DBDuplicateEntryError(
-            "Duplicate entry for Action ['name', 'project_id']:"
-            " {}, {}".format(a_def.name, a_def.project_id)
+            "Duplicate entry for Action ['name', 'namespace', 'project_id']:"
+            " {}, {}, {}".format(a_def.name, a_def.namespace, a_def.project_id)
         )
 
     return a_def
@@ -684,7 +712,8 @@ def create_action_definition(values, session=None):
 
 @b.session_aware()
 def update_action_definition(identifier, values, session=None):
-    a_def = get_action_definition(identifier)
+    namespace = values.get('namespace', '')
+    a_def = get_action_definition(identifier, namespace=namespace)
 
     a_def.update(values.copy())
 
@@ -693,15 +722,19 @@ def update_action_definition(identifier, values, session=None):
 
 @b.session_aware()
 def create_or_update_action_definition(name, values, session=None):
-    if not _get_db_object_by_name(models.ActionDefinition, name):
+    namespace = values.get('namespace', '')
+    if not _get_db_object_by_name_and_namespace(
+            models.ActionDefinition,
+            name,
+            namespace=namespace):
         return create_action_definition(values)
     else:
         return update_action_definition(name, values)
 
 
 @b.session_aware()
-def delete_action_definition(identifier, session=None):
-    a_def = get_action_definition(identifier)
+def delete_action_definition(identifier, namespace='', session=None):
+    a_def = get_action_definition(identifier, namespace=namespace)
 
     session.delete(a_def)
 
@@ -769,6 +802,17 @@ def create_or_update_action_execution(id, values, session=None):
         return update_action_execution(id, values)
 
 
+@b.session_aware()
+def update_action_execution_heartbeat(id, session=None):
+    if not id:
+        raise exc.DBEntityNotFoundError
+
+    now = utils.utc_now_sec()
+    session.query(models.ActionExecution). \
+        filter(models.ActionExecution.id == id). \
+        update({'last_heartbeat': now})
+
+
 @b.session_aware()
 def delete_action_execution(id, session=None):
     count = _secure_query(models.ActionExecution).filter(
@@ -863,13 +907,59 @@ def delete_workflow_execution(id, session=None):
     insecure = context.ctx().is_admin
     query = b.model_query(model) if insecure else _secure_query(model)
 
-    count = query.filter(
-        models.WorkflowExecution.id == id).delete()
+    try:
+        count = query.filter(
+            models.WorkflowExecution.id == id
+        ).delete()
 
-    if count == 0:
-        raise exc.DBEntityNotFoundError(
-            "WorkflowExecution not found [id=%s]" % id
-        )
+        if count == 0:
+            raise exc.DBEntityNotFoundError(
+                "WorkflowExecution not found [id=%s]" % id
+            )
+    except Exception as e:
+        if is_mysql_max_depth_error(e) or is_mariadb_max_depth_error(e):
+            # https://bugs.launchpad.net/mistral/+bug/1832300
+            # mysql cascade delete error
+            delete_workflow_execution_recurse(id)
+        else:
+            raise
+
+
+def is_mysql_max_depth_error(e):
+    pattern = ".*3008.*Foreign key cascade delete" \
+              "/update exceeds max depth of 15.*"
+    return re.match(pattern, str(e))
+
+
+def is_mariadb_max_depth_error(e):
+    pattern = ".*Got error 193.*ON DELETE CASCADE.*"
+    return re.match(pattern, str(e))
+
+
+def delete_workflow_execution_recurse(wf_ex_id):
+    sub_wf_ex_ids = _get_all_direct_subworkflows(wf_ex_id)
+
+    for sub_wf_ex_id in sub_wf_ex_ids:
+        delete_workflow_execution(sub_wf_ex_id)
+
+    delete_workflow_execution(wf_ex_id)
+
+
+def _get_all_direct_subworkflows(wf_ex_id):
+    model = models.WorkflowExecution
+    insecure = context.ctx().is_admin
+    if insecure:
+        query = b.model_query(model, [model.id])
+    else:
+        query = _secure_query(model, model.id)
+    query = query.join(
+        models.TaskExecution,
+        models.WorkflowExecution.task_execution_id == models.TaskExecution.id
+    ).filter(
+        models.TaskExecution.workflow_execution_id == wf_ex_id
+    )
+
+    return [i[0] for i in query.all()]
 
 
 @b.session_aware()
@@ -907,6 +997,15 @@ def get_task_executions(session=None, **kwargs):
     return _get_collection(models.TaskExecution, **kwargs)
 
 
+@b.session_aware()
+def get_task_executions_count(session=None, **kwargs):
+    query = b.model_query(models.TaskExecution)
+
+    query = query.filter_by(**kwargs)
+
+    return query.count()
+
+
 def _get_completed_task_executions_query(kwargs):
     query = b.model_query(models.TaskExecution)
 
@@ -1127,11 +1226,140 @@ def get_delayed_calls(session=None, **kwargs):
     return _get_collection(model=models.DelayedCall, **kwargs)
 
 
+@b.session_aware()
+def get_delayed_calls_count(session=None, **kwargs):
+    return _get_count(model=models.DelayedCall, **kwargs)
+
+
 @b.session_aware()
 def delete_delayed_calls(session=None, **kwargs):
     return _delete_all(models.DelayedCall, **kwargs)
 
 
+# Scheduled jobs.
+
+@b.session_aware()
+def create_scheduled_job(values, session=None):
+    job = models.ScheduledJob()
+
+    job.update(values.copy())
+
+    try:
+        job.save(session)
+    except db_exc.DBDuplicateEntry as e:
+        raise exc.DBDuplicateEntryError(
+            "Duplicate entry for ScheduledJob ID: {}".format(e.value)
+        )
+
+    return job
+
+
+@b.session_aware()
+def get_scheduled_jobs_to_start(time, batch_size=None, session=None):
+    query = b.model_query(models.ScheduledJob)
+
+    execute_at_col = models.ScheduledJob.execute_at
+    captured_at_col = models.ScheduledJob.captured_at
+
+    # Filter by execution time accounting for a configured job pickup interval.
+    # TODO(rakhmerov): Configuration options should not be accessed here.
+    query = query.filter(
+        execute_at_col <
+        time - datetime.timedelta(seconds=CONF.scheduler.pickup_job_after)
+    )
+
+    # Filter by captured time accounting for a configured captured job timeout.
+    min_captured_at = (
+        utils.utc_now_sec() -
+        datetime.timedelta(seconds=CONF.scheduler.captured_job_timeout)
+    )
+
+    query = query.filter(
+        sa.or_(
+            captured_at_col == sa.null(),
+            captured_at_col <= min_captured_at
+        )
+    )
+
+    query = query.order_by(execute_at_col)
+    query = query.limit(batch_size)
+
+    return query.all()
+
+
+@b.session_aware()
+def update_scheduled_job(id, values, query_filter=None, session=None):
+    if query_filter:
+        try:
+            specimen = models.ScheduledJob(id=id, **query_filter)
+
+            job = b.model_query(
+                models.ScheduledJob
+            ).update_on_match(
+                specimen=specimen,
+                surrogate_key='id',
+                values=values
+            )
+
+            return job, 1
+
+        except oslo_sqlalchemy.update_match.NoRowsMatched as e:
+            LOG.debug(
+                "No rows matched for update scheduled job [id=%s, values=%s, "
+                "query_filter=%s,"
+                "exception=%s]", id, values, query_filter, e
+            )
+
+            return None, 0
+
+    else:
+        job = get_scheduled_job(id=id, session=session)
+
+        job.update(values)
+
+        return job, len(session.dirty)
+
+
+@b.session_aware()
+def get_scheduled_job(id, session=None):
+    job = _get_db_object_by_id(models.ScheduledJob, id)
+
+    if not job:
+        raise exc.DBEntityNotFoundError(
+            "Scheduled job not found [id=%s]" % id
+        )
+
+    return job
+
+
+@b.session_aware()
+def delete_scheduled_job(id, session=None):
+    # It's safe to use insecure query here because users can't access
+    # scheduled job.
+    count = b.model_query(models.ScheduledJob).filter(
+        models.ScheduledJob.id == id).delete()
+
+    if count == 0:
+        raise exc.DBEntityNotFoundError(
+            "Scheduled job not found [id=%s]" % id
+        )
+
+
+def get_scheduled_jobs(**kwargs):
+    return _get_collection(model=models.ScheduledJob, **kwargs)
+
+
+@b.session_aware()
+def delete_scheduled_jobs(session=None, **kwargs):
+    return _delete_all(models.ScheduledJob, **kwargs)
+
+
+def get_scheduled_jobs_count(**kwargs):
+    return _get_count(model=models.ScheduledJob, **kwargs)
+
+
+# Other functions.
+
 @b.session_aware()
 def get_expired_executions(expiration_time, limit=None, columns=(),
                            session=None):
@@ -1145,14 +1373,19 @@ def get_expired_executions(expiration_time, limit=None, columns=(),
 
 
 @b.session_aware()
-def get_running_expired_sync_actions(expiration_time, session=None):
+def get_running_expired_sync_action_executions(expiration_time,
+                                               limit, session=None):
     query = b.model_query(models.ActionExecution)
+
     query = query.filter(
         models.ActionExecution.last_heartbeat < expiration_time
     )
     query = query.filter_by(is_sync=True)
     query = query.filter(models.ActionExecution.state == states.RUNNING)
 
+    if limit:
+        query.limit(limit)
+
     return query.all()
 
 
@@ -1175,17 +1408,18 @@ def get_superfluous_executions(max_finished_executions, limit=None, columns=(),
 def _get_completed_root_executions_query(columns):
     query = b.model_query(models.WorkflowExecution, columns=columns)
 
-    # Only WorkflowExecution that are not a child of other WorkflowExecution.
+    # This is an empty list by default.
+    ignored_states = CONF.execution_expiration_policy.ignored_states
+    desired_states = states.TERMINAL_STATES - set(ignored_states)
+
+    # Only workflow executions that are not a child of
+    # other workflow executions.
     query = query.filter(
         models.WorkflowExecution.task_execution_id == sa.null()
     )
 
     query = query.filter(
-        models.WorkflowExecution.state.in_(
-            [states.SUCCESS,
-             states.ERROR,
-             states.CANCELLED]
-        )
+        models.WorkflowExecution.state.in_(desired_states)
     )
 
     return query
@@ -1539,8 +1773,8 @@ def update_resource_member(resource_id, res_type, member_id, values,
 
 @b.session_aware()
 def delete_resource_member(resource_id, res_type, member_id, session=None):
-    query = _secure_query(models.ResourceMember).\
-        filter_by(resource_type=res_type).\
+    query = _secure_query(models.ResourceMember). \
+        filter_by(resource_type=res_type). \
         filter(_get_criterion(resource_id, member_id))
 
     # TODO(kong): Check association with cron triggers when deleting a workflow
