@@ -1,5 +1,6 @@
 # Copyright 2015 - Mirantis, Inc.
 # Copyright 2015 - StackStorm, Inc.
+# Copyright 2020 Nokia Software.
 #
 #    Licensed under the Apache License, Version 2.0 (the "License");
 #    you may not use this file except in compliance with the License.
@@ -29,8 +30,7 @@ from mistral.db.sqlalchemy import model_base as mb
 from mistral.db.sqlalchemy import types as st
 from mistral import exceptions as exc
 from mistral.services import security
-from mistral import utils
-
+from mistral_lib import utils
 
 # Definition objects.
 
@@ -82,6 +82,19 @@ def validate_long_type_length(cls, field_name, value):
             raise exc.SizeLimitExceededException(msg)
 
 
+def validate_name_has_no_spaces(name):
+    """Makes sure name does not contain spaces."""
+    if name:
+        if " " in name:
+            msg = (
+                "Name '{}' must not contain spaces"
+            ).format(name)
+
+            LOG.error(msg)
+
+            raise exc.InvalidModelException(msg)
+
+
 def register_length_validator(attr_name):
     """Register an event listener on the attribute.
 
@@ -97,11 +110,26 @@ def register_length_validator(attr_name):
             )
 
 
+def register_name_validator():
+    """Register an event listener on the attribute.
+
+    This event listener will validate that name of object does not
+    contains spaces every time a 'set' occurs.
+    """
+    for cls in utils.iter_subclasses(Definition):
+        event.listen(
+            getattr(cls, "name"),
+            'set',
+            lambda t, v, o, i: validate_name_has_no_spaces(v)
+        )
+
+
 class Definition(mb.MistralSecureModelBase):
     __abstract__ = True
 
     id = mb.id_column()
     name = sa.Column(sa.String(255))
+    namespace = sa.Column(sa.String(255), nullable=True)
     definition = sa.Column(st.MediumText(), nullable=True)
     spec = sa.Column(st.JsonMediumDictType())
     tags = sa.Column(st.JsonListType())
@@ -113,8 +141,6 @@ class Workbook(Definition):
     """Contains info about workbook (including definition in Mistral DSL)."""
 
     __tablename__ = 'workbooks_v2'
-    namespace = sa.Column(sa.String(255), nullable=True)
-
     __table_args__ = (
         sa.UniqueConstraint(
             'name',
@@ -130,7 +156,6 @@ class WorkflowDefinition(Definition):
     """Contains info about workflow (including definition in Mistral DSL)."""
 
     __tablename__ = 'workflow_definitions_v2'
-    namespace = sa.Column(sa.String(255), nullable=True)
     __table_args__ = (
         sa.UniqueConstraint(
             'name',
@@ -142,20 +167,26 @@ class WorkflowDefinition(Definition):
         sa.Index('%s_scope' % __tablename__, 'scope'),
     )
 
+    workbook_name = sa.Column(sa.String(255))
+
 
 class ActionDefinition(Definition):
     """Contains info about registered Actions."""
 
     __tablename__ = 'action_definitions_v2'
-
     __table_args__ = (
-        sa.UniqueConstraint('name', 'project_id'),
+        sa.UniqueConstraint(
+            'name',
+            'namespace',
+            'project_id'),
         sa.Index('%s_is_system' % __tablename__, 'is_system'),
         sa.Index('%s_action_class' % __tablename__, 'action_class'),
         sa.Index('%s_project_id' % __tablename__, 'project_id'),
         sa.Index('%s_scope' % __tablename__, 'scope'),
     )
 
+    workbook_name = sa.Column(sa.String(255))
+
     # Main properties.
     description = sa.Column(sa.Text())
     input = sa.Column(sa.Text())
@@ -177,7 +208,6 @@ class Execution(mb.MistralSecureModelBase):
     workflow_name = sa.Column(sa.String(255))
     workflow_namespace = sa.Column(sa.String(255))
     workflow_id = sa.Column(sa.String(80))
-    spec = sa.Column(st.JsonMediumDictType())
     state = sa.Column(sa.String(20))
     state_info = sa.Column(sa.Text(), nullable=True)
     tags = sa.Column(st.JsonListType())
@@ -199,6 +229,7 @@ class ActionExecution(Execution):
     )
 
     # Main properties.
+    spec = sa.Column(st.JsonMediumDictType())
     accepted = sa.Column(sa.Boolean(), default=False)
     input = sa.Column(st.JsonLongDictType(), nullable=True)
     output = sa.orm.deferred(sa.Column(st.JsonLongDictType(), nullable=True))
@@ -224,10 +255,11 @@ class WorkflowExecution(Execution):
     )
 
     # Main properties.
+    spec = sa.orm.deferred(sa.Column(st.JsonMediumDictType()))
     accepted = sa.Column(sa.Boolean(), default=False)
-    input = sa.Column(st.JsonLongDictType(), nullable=True)
+    input = sa.orm.deferred(sa.Column(st.JsonLongDictType(), nullable=True))
     output = sa.orm.deferred(sa.Column(st.JsonLongDictType(), nullable=True))
-    params = sa.Column(st.JsonLongDictType())
+    params = sa.orm.deferred(sa.Column(st.JsonLongDictType()))
 
     # Initial workflow context containing workflow variables, environment,
     # openstack security context etc.
@@ -235,7 +267,7 @@ class WorkflowExecution(Execution):
     #   * Data stored in this structure should not be copied into inbound
     #     contexts of tasks. No need to duplicate it.
     #   * This structure does not contain workflow input.
-    context = sa.Column(st.JsonLongDictType())
+    context = sa.orm.deferred(sa.Column(st.JsonLongDictType()))
 
 
 class TaskExecution(Execution):
@@ -252,15 +284,35 @@ class TaskExecution(Execution):
     )
 
     # Main properties.
+    spec = sa.orm.deferred(sa.Column(st.JsonMediumDictType()))
     action_spec = sa.Column(st.JsonLongDictType())
     unique_key = sa.Column(sa.String(255), nullable=True)
     type = sa.Column(sa.String(10))
+    started_at = sa.Column(sa.DateTime, nullable=True)
+    finished_at = sa.Column(sa.DateTime, nullable=True)
 
     # Whether the task is fully processed (publishing and calculating commands
     # after it). It allows to simplify workflow controller implementations
     # significantly.
     processed = sa.Column(sa.BOOLEAN, default=False)
 
+    # Set to True if the completion of the task led to starting new
+    # tasks.
+    # The value of this property should be ignored if the task
+    # is not completed.
+    has_next_tasks = sa.Column(sa.Boolean, default=False)
+
+    # The names of the next tasks.
+    # [(task_name, event)]
+    next_tasks = sa.Column(st.JsonListType())
+
+    # Set to True if the task finished with an error and the error
+    # is handled (e.g. with 'on-error' clause for direct workflows)
+    # so that the error shouldn't bubble up to the workflow level.
+    # The value of this property should be ignored if the task
+    # is not completed.
+    error_handled = sa.Column(sa.Boolean, default=False)
+
     # Data Flow properties.
     in_context = sa.Column(st.JsonLongDictType())
     published = sa.Column(st.JsonLongDictType())
@@ -273,6 +325,14 @@ class TaskExecution(Execution):
             else self.workflow_executions
         )
 
+    def to_dict(self):
+        d = super(TaskExecution, self).to_dict()
+
+        utils.datetime_to_str_in_dict(d, 'started_at')
+        utils.datetime_to_str_in_dict(d, 'finished_at')
+
+        return d
+
 
 for cls in utils.iter_subclasses(Execution):
     event.listen(
@@ -290,7 +350,6 @@ for cls in utils.iter_subclasses(Execution):
         retval=True
     )
 
-
 # Many-to-one for 'ActionExecution' and 'TaskExecution'.
 
 ActionExecution.task_execution_id = sa.Column(
@@ -343,6 +402,12 @@ WorkflowExecution.root_execution_id = sa.Column(
     nullable=True
 )
 
+WorkflowExecution.root_execution = relationship(
+    WorkflowExecution,
+    remote_side=WorkflowExecution.id,
+    lazy='select'
+)
+
 # Many-to-one for 'TaskExecution' and 'WorkflowExecution'.
 
 TaskExecution.workflow_execution_id = sa.Column(
@@ -384,7 +449,7 @@ class DelayedCall(mb.MistralModelBase):
     method_arguments = sa.Column(st.JsonDictType())
     serializers = sa.Column(st.JsonDictType())
     key = sa.Column(sa.String(250), nullable=True)
-    auth_context = sa.Column(st.JsonDictType())
+    auth_context = sa.Column(st.JsonMediumDictType())
     execution_time = sa.Column(sa.DateTime, nullable=False)
     processing = sa.Column(sa.Boolean, default=False, nullable=False)
 
@@ -395,6 +460,35 @@ sa.Index(
 )
 
 
+class ScheduledJob(mb.MistralModelBase):
+    """Contains info about scheduled jobs."""
+
+    __tablename__ = 'scheduled_jobs_v2'
+
+    id = mb.id_column()
+
+    run_after = sa.Column(sa.Integer)
+
+    # The full name of the factory function that returns/builds a Python
+    # (target) object whose method should be called. Optional.
+    target_factory_func_name = sa.Column(sa.String(200), nullable=True)
+
+    # May take two different forms:
+    # 1. Full path of a target function that should be called. For example,
+    #  "mistral.utils.random_sleep".
+    # 2. Name of a method to call on a target object, if
+    #  "target_factory_func_name" is specified.
+    func_name = sa.Column(sa.String(80), nullable=False)
+
+    func_args = sa.Column(st.JsonDictType())
+    func_arg_serializers = sa.Column(st.JsonDictType())
+    auth_ctx = sa.Column(st.JsonDictType())
+    execute_at = sa.Column(sa.DateTime, nullable=False)
+    captured_at = sa.Column(sa.DateTime, nullable=True)
+
+    key = sa.Column(sa.String(250), nullable=True)
+
+
 class Environment(mb.MistralSecureModelBase):
     """Contains environment variables for workflow execution."""
 
@@ -485,6 +579,8 @@ mb.register_secure_model_hooks()
 for attr_name in ['input', 'output', 'params', 'published']:
     register_length_validator(attr_name)
 
+register_name_validator()
+
 
 class ResourceMember(mb.MistralModelBase):
     """Contains info about resource members."""
