@@ -1,6 +1,7 @@
 # Copyright 2016 - Nokia Networks.
 # Copyright 2016 - Brocade Communications Systems, Inc.
 # Copyright 2018 - Extreme Networks, Inc.
+# Copyright 2019 - NetCracker Technology Corp.
 #
 #    Licensed under the Apache License, Version 2.0 (the "License");
 #    you may not use this file except in compliance with the License.
@@ -15,33 +16,35 @@
 #    limitations under the License.
 
 import abc
+import collections
 import copy
 import json
 from oslo_config import cfg
 from oslo_log import log as logging
 from osprofiler import profiler
-import six
 
 from mistral.db.v2 import api as db_api
 from mistral.engine import actions
 from mistral.engine import dispatcher
 from mistral.engine import policies
+from mistral.engine import post_tx_queue
+from mistral.engine import workflow_handler as wf_handler
 from mistral import exceptions as exc
 from mistral import expressions as expr
 from mistral.notifiers import base as notif
 from mistral.notifiers import notification_events as events
-from mistral import utils
+from mistral.services import actions as action_service
 from mistral.utils import wf_trace
 from mistral.workflow import base as wf_base
+from mistral.workflow import commands
 from mistral.workflow import data_flow
 from mistral.workflow import states
-
+from mistral_lib import utils
 
 LOG = logging.getLogger(__name__)
 
 
-@six.add_metaclass(abc.ABCMeta)
-class Task(object):
+class Task(object, metaclass=abc.ABCMeta):
     """Task.
 
     Represents a workflow task and defines interface that can be used by
@@ -49,7 +52,8 @@ class Task(object):
     """
 
     def __init__(self, wf_ex, wf_spec, task_spec, ctx, task_ex=None,
-                 unique_key=None, waiting=False, triggered_by=None):
+                 unique_key=None, waiting=False, triggered_by=None,
+                 rerun=False):
         self.wf_ex = wf_ex
         self.task_spec = task_spec
         self.ctx = ctx
@@ -58,26 +62,74 @@ class Task(object):
         self.unique_key = unique_key
         self.waiting = waiting
         self.triggered_by = triggered_by
+        self.rerun = rerun
         self.reset_flag = False
         self.created = False
         self.state_changed = False
 
-    def notify(self, old_task_state, new_task_state):
+    def _notify(self, from_state, to_state):
         publishers = self.wf_ex.params.get('notify')
 
         if not publishers and not isinstance(publishers, list):
             return
 
         notifier = notif.get_notifier(cfg.CONF.notifier.type)
-        event = events.identify_task_event(old_task_state, new_task_state)
 
-        notifier.notify(
-            self.task_ex.id,
-            self.task_ex.to_dict(),
-            event,
-            self.task_ex.updated_at,
-            publishers
-        )
+        event = events.identify_task_event(from_state, to_state)
+
+        filtered_publishers = []
+
+        for publisher in publishers:
+            if not isinstance(publisher, dict):
+                continue
+
+            target_events = publisher.get('event_types', [])
+
+            if not target_events or event in target_events:
+                filtered_publishers.append(publisher)
+
+        if not filtered_publishers:
+            return
+
+        data = {
+            "id": self.task_ex.id,
+            "name": self.task_ex.name,
+            "workflow_execution_id": self.task_ex.workflow_execution_id,
+            "workflow_name": self.task_ex.workflow_name,
+            "workflow_namespace": self.task_ex.workflow_namespace,
+            "workflow_id": self.task_ex.workflow_id,
+            "state": self.task_ex.state,
+            "state_info": self.task_ex.state_info,
+            "type": self.task_ex.type,
+            "project_id": self.task_ex.project_id,
+            "created_at": utils.datetime_to_str(self.task_ex.created_at),
+            "updated_at": utils.datetime_to_str(self.task_ex.updated_at),
+            "started_at": utils.datetime_to_str(self.task_ex.started_at),
+            "finished_at": utils.datetime_to_str(self.task_ex.finished_at)
+        }
+
+        def _send_notification():
+            notifier.notify(
+                self.task_ex.id,
+                data,
+                event,
+                self.task_ex.updated_at,
+                filtered_publishers
+            )
+
+        post_tx_queue.register_operation(_send_notification)
+
+    def get_id(self):
+        return self.task_ex.id if self.task_ex else None
+
+    def get_name(self):
+        return self.task_ex.name if self.task_ex else None
+
+    def get_state(self):
+        return self.task_ex.state if self.task_ex else None
+
+    def get_state_info(self):
+        return self.task_ex.state_info if self.task_ex else None
 
     def is_completed(self):
         return self.task_ex and states.is_completed(self.task_ex.state)
@@ -91,6 +143,85 @@ class Task(object):
     def is_state_changed(self):
         return self.state_changed
 
+    def get_expression_context(self, ctx=None):
+        assert self.task_ex
+
+        return data_flow.ContextView(
+            data_flow.get_current_task_dict(self.task_ex),
+            data_flow.get_workflow_environment_dict(self.wf_ex),
+            ctx or {},
+            self.task_ex.in_context,
+            self.wf_ex.context,
+            self.wf_ex.input,
+        )
+
+    def evaluate(self, data, ctx=None):
+        """Evaluates data against the task context.
+
+        The data is evaluated against the standard task context that includes
+        all standard components like workflow environment, workflow input etc.
+        However, if needed, the method also takes an additional context that
+        can be virtually merged with the standard one.
+
+        :param data: Data (a string, dict or a list) that possibly contain
+            YAQL/Jinja expressions to evaluate.
+        :param ctx: Additional context.
+        :return:
+        """
+
+        return expr.evaluate_recursively(
+            data,
+            self.get_expression_context(ctx)
+        )
+
+    def set_runtime_context_value(self, key, value):
+        assert self.task_ex
+
+        if not self.task_ex.runtime_context:
+            self.task_ex.runtime_context = {}
+
+        self.task_ex.runtime_context[key] = value
+
+    def touch_runtime_context(self):
+        """Must be called after any update of the runtime context.
+
+
+        The ORM framework can't trace updates happening within
+        deep structures like dictionaries. So every time we update
+        something inside "runtime_context" of the task execution
+        we need to call this method that does a fake update of the
+        field.
+        """
+        runtime_ctx = self.task_ex.runtime_context
+
+        if runtime_ctx:
+            random_key = next(iter(runtime_ctx.keys()))
+            runtime_ctx[random_key] = runtime_ctx[random_key]
+
+    def cleanup_runtime_context(self):
+        runtime_context = self.task_ex.runtime_context
+
+        if runtime_context:
+            runtime_context.clear()
+
+    def get_policy_context(self, key):
+        assert self.task_ex
+
+        if not self.task_ex.runtime_context:
+            self.task_ex.runtime_context = {}
+
+        if key not in self.task_ex.runtime_context:
+            self.task_ex.runtime_context.update({key: {}})
+
+        return self.task_ex.runtime_context[key]
+
+    def invalidate_result(self):
+        if not self.task_ex:
+            return
+
+        for ex in self.task_ex.executions:
+            ex.accepted = False
+
     @abc.abstractmethod
     def on_action_complete(self, action_ex):
         """Handle action completion.
@@ -118,6 +249,23 @@ class Task(object):
 
         This method puts task to a waiting state.
         """
+
+        # NOTE(rakhmerov): using named locks may cause problems under load
+        # with MySQL that raises a lot of deadlocks in case of high
+        # parallelism so it makes sense to do a fast check if the object
+        # already exists in DB outside of the lock.
+        if not self.task_ex:
+            t_execs = db_api.get_task_executions(
+                workflow_execution_id=self.wf_ex.id,
+                unique_key=self.unique_key,
+                state=states.WAITING
+            )
+
+            self.task_ex = t_execs[0] if t_execs else None
+
+        if self.task_ex:
+            return
+
         with db_api.named_lock(self.unique_key):
             if not self.task_ex:
                 t_execs = db_api.get_task_executions(
@@ -171,9 +319,25 @@ class Task(object):
                 if isinstance(state_info, dict) else state_info
             self.state_changed = True
 
+            # Recalculating "started_at" timestamp only if the state
+            # was WAITING (all preconditions are satisfied and it's
+            # ready to start) or IDLE, or the task is being rerun. So
+            # we treat all iterations of "retry" policy as one run.
+            if state == states.RUNNING and \
+                    (cur_state in (None, states.WAITING) or self.rerun):
+                self.task_ex.started_at = utils.utc_now_sec()
+
+            if states.is_completed(state):
+                self.task_ex.finished_at = utils.utc_now_sec()
+
+            if self.rerun:
+                self.task_ex.finished_at = None
+
             if processed is not None:
                 self.task_ex.processed = processed
 
+            self._notify(cur_state, state)
+
             wf_trace.info(
                 self.task_ex.workflow_execution,
                 "Task '%s' (%s) [%s -> %s, msg=%s]" %
@@ -200,15 +364,8 @@ class Task(object):
 
         assert self.task_ex
 
-        # Record the current task state.
-        old_task_state = self.task_ex.state
-
         # Ignore if task already completed.
         if self.is_completed():
-            # Publish task event again so subscribers know
-            # task completed state is being processed again.
-            self.notify(old_task_state, self.task_ex.state)
-
             return
 
         # If we were unable to change the task state it means that it was
@@ -232,28 +389,54 @@ class Task(object):
         if self.task_ex.state == states.RUNNING_DELAYED:
             return
 
-        # If workflow is paused we shouldn't schedule new commands
-        # and mark task as processed.
-        if states.is_paused(self.wf_ex.state):
-            # Publish task event even if the workflow is paused.
-            self.notify(old_task_state, self.task_ex.state)
-
-            return
-
         wf_ctrl = wf_base.get_controller(self.wf_ex, self.wf_spec)
 
         # Calculate commands to process next.
         cmds = wf_ctrl.continue_workflow(task_ex=self.task_ex)
 
+        # Save next task names in DB to avoid evaluating them again
+        # in the future.
+        self.task_ex.next_tasks = []
+
+        for c in cmds:
+            if commands.is_engine_command(c):
+                continue
+
+            event = c.triggered_by[0]['event'] if c.triggered_by else None
+
+            self.task_ex.next_tasks.append((c.task_spec.get_name(), event))
+
+        self.task_ex.has_next_tasks = bool(self.task_ex.next_tasks)
+
+        # Check whether the error is handled.
+        if self.task_ex.state == states.ERROR:
+            self.task_ex.error_handled = any([c.handles_error for c in cmds])
+
+        # If workflow is paused we shouldn't schedule new commands
+        # and mark task as processed.
+        if states.is_paused(self.wf_ex.state):
+            return
+
         # Mark task as processed after all decisions have been made
         # upon its completion.
         self.task_ex.processed = True
 
-        # Publish task event.
-        self.notify(old_task_state, self.task_ex.state)
+        self.register_workflow_completion_check()
 
         dispatcher.dispatch_workflow_commands(self.wf_ex, cmds)
 
+    def register_workflow_completion_check(self):
+        wf_ctrl = wf_base.get_controller(self.wf_ex, self.wf_spec)
+
+        # Register an asynchronous command to check workflow completion
+        # in a separate transaction if the task may potentially lead to
+        # workflow completion.
+        def _check():
+            wf_handler.check_and_complete(self.wf_ex.id)
+
+        if wf_ctrl.may_complete_workflow(self.task_ex):
+            post_tx_queue.register_operation(_check, in_tx=True)
+
     @profiler.trace('task-update')
     def update(self, state, state_info=None):
         """Update task and set specified state.
@@ -266,15 +449,8 @@ class Task(object):
 
         assert self.task_ex
 
-        # Record the current task state.
-        old_task_state = self.task_ex.state
-
         # Ignore if task already completed.
         if states.is_completed(self.task_ex.state):
-            # Publish task event again so subscribers know
-            # task completed state is being processed again.
-            self.notify(old_task_state, self.task_ex.state)
-
             return
 
         # Update only if state transition is valid.
@@ -290,34 +466,31 @@ class Task(object):
 
         self.set_state(state, state_info)
 
-        # Publish event.
-        self.notify(old_task_state, self.task_ex.state)
+        if states.is_completed(self.task_ex.state):
+            self.register_workflow_completion_check()
 
     def _before_task_start(self):
         policies_spec = self.task_spec.get_policies()
 
         for p in policies.build_policies(policies_spec, self.wf_spec):
-            p.before_task_start(self.task_ex, self.task_spec)
+            p.before_task_start(self)
 
     def _after_task_complete(self):
         policies_spec = self.task_spec.get_policies()
 
         for p in policies.build_policies(policies_spec, self.wf_spec):
-            p.after_task_complete(self.task_ex, self.task_spec)
+            p.after_task_complete(self)
 
     @profiler.trace('task-create-task-execution')
-    def _create_task_execution(self, state=states.RUNNING, state_info=None):
-        task_id = utils.generate_unicode_uuid()
-        task_name = self.task_spec.get_name()
-        task_type = self.task_spec.get_type()
-
+    def _create_task_execution(self, state=None, state_info=None):
         values = {
-            'id': task_id,
-            'name': task_name,
+            'id': utils.generate_unicode_uuid(),
+            'name': self.task_spec.get_name(),
             'workflow_execution_id': self.wf_ex.id,
             'workflow_name': self.wf_ex.workflow_name,
             'workflow_namespace': self.wf_ex.workflow_namespace,
             'workflow_id': self.wf_ex.workflow_id,
+            'tags': self.task_spec.get_tags(),
             'state': state,
             'state_info': state_info,
             'spec': self.task_spec.to_dict(),
@@ -326,7 +499,7 @@ class Task(object):
             'published': {},
             'runtime_context': {},
             'project_id': self.wf_ex.project_id,
-            'type': task_type
+            'type': self.task_spec.get_type(),
         }
 
         if self.triggered_by:
@@ -375,8 +548,13 @@ class RegularTask(Task):
         # TODO(rakhmerov): Here we can define more informative messages for
         # cases when action is successful and when it's not. For example,
         # in state_info we can specify the cause action.
-        state_info = (None if state == states.SUCCESS
-                      else action_ex.output.get('result'))
+
+        if state == states.SUCCESS:
+            state_info = None
+        else:
+            action_result = action_ex.output.get('result')
+
+            state_info = str(action_result) if action_result else None
 
         self.complete(state, state_info)
 
@@ -400,14 +578,17 @@ class RegularTask(Task):
 
         self._create_task_execution()
 
-        # Publish event.
-        self.notify(None, self.task_ex.state)
+        # Set the initial state and trigger all operations
+        # related to the state change.
+        self.set_state(states.RUNNING, None)
 
         LOG.debug(
-            'Starting task [workflow=%s, task=%s, init_state=%s]',
-            self.wf_ex.name,
+            'Starting task [name=%s, init_state=%s, workflow_name=%s,'
+            ' execution_id=%s]',
             self.task_spec.get_name(),
-            self.task_ex.state
+            self.task_ex.state,
+            self.wf_ex.name,
+            self.wf_ex.id
         )
 
         self._before_task_start()
@@ -430,13 +611,14 @@ class RegularTask(Task):
                 'Rerunning succeeded tasks is not supported.'
             )
 
-        # Record the current task state.
-        old_task_state = self.task_ex.state
-
         self.set_state(states.RUNNING, None, processed=False)
 
-        # Publish event.
-        self.notify(old_task_state, self.task_ex.state)
+        if self.rerun:
+            self._before_task_start()
+
+            # Policies could possibly change task state.
+            if self.task_ex.state != states.RUNNING:
+                return
 
         self._update_inbound_context()
         self._update_triggered_by()
@@ -485,8 +667,6 @@ class RegularTask(Task):
 
         action = self._build_action()
 
-        action.validate_input(input_dict)
-
         action.schedule(
             input_dict,
             target,
@@ -496,6 +676,9 @@ class RegularTask(Task):
 
     @profiler.trace('regular-task-get-target', hide_args=True)
     def _get_target(self, input_dict):
+        if not self.task_spec.get_target():
+            return None
+
         ctx_view = data_flow.ContextView(
             input_dict,
             self.ctx,
@@ -511,7 +694,11 @@ class RegularTask(Task):
 
     @profiler.trace('regular-task-get-action-input', hide_args=True)
     def _get_action_input(self, ctx=None):
-        input_dict = self._evaluate_expression(self.task_spec.get_input(), ctx)
+        input_spec = self.task_spec.get_input()
+
+        input_dict = (
+            self.evaluate(input_spec, ctx) if input_spec else {}
+        )
 
         if not isinstance(input_dict, dict):
             raise exc.InputException(
@@ -526,50 +713,66 @@ class RegularTask(Task):
             overwrite=False
         )
 
-    def _evaluate_expression(self, expression, ctx=None):
-        ctx_view = data_flow.ContextView(
-            data_flow.get_current_task_dict(self.task_ex),
-            data_flow.get_workflow_environment_dict(self.wf_ex),
-            ctx or self.ctx,
-            self.wf_ex.context,
-            self.wf_ex.input
-        )
+    def _get_action_name(self):
+        result = self.task_spec.get_action_name()
 
-        return expr.evaluate_recursively(
-            expression,
-            ctx_view
-        )
+        # An action name can be an expression so we reevaluate it.
+        if result:
+            result = self.evaluate(result)
+
+        if not result:
+            result = 'std.noop'
+
+        return result
 
     def _build_action(self):
-        action_name = self.task_spec.get_action_name()
         wf_name = self.task_spec.get_workflow_name()
 
-        # For dynamic workflow evaluation we regenerate the action.
+        # A workflow name can be an expression so we reevaluate it.
         if wf_name:
             return actions.WorkflowAction(
-                wf_name=self._evaluate_expression(wf_name),
+                wf_name=self.evaluate(wf_name),
                 task_ex=self.task_ex
             )
 
-        # For dynamic action evaluation we just regenerate the name.
-        if action_name:
-            action_name = self._evaluate_expression(action_name)
+        action_desc = self._get_action_descriptor()
 
-        if not action_name:
-            action_name = 'std.noop'
-
-        action_def = actions.resolve_action_definition(
-            action_name,
-            self.wf_ex.name,
-            self.wf_spec.get_name()
+        return actions.RegularAction(
+            action_desc=action_desc,
+            task_ex=self.task_ex,
+            task_ctx=self.ctx
         )
 
-        if action_def.spec:
-            return actions.AdHocAction(action_def, task_ex=self.task_ex,
-                                       task_ctx=self.ctx,
-                                       wf_ctx=self.wf_ex.context)
+    def _get_action_descriptor(self):
+        res = None
 
-        return actions.PythonAction(action_def, task_ex=self.task_ex)
+        action_name = self._get_action_name()
+        namespace = self.wf_ex.workflow_namespace
+
+        provider = action_service.get_system_action_provider()
+
+        wb_name = self.wf_ex.runtime_context.get('wb_name')
+
+        if wb_name:
+            # First try to find the action within the same workbook
+            # that the workflow belongs to.
+            res = provider.find(
+                '%s.%s' % (wb_name, action_name),
+                namespace=namespace
+            )
+
+        if res is None:
+            # Now try to find by the action name as it appears in the
+            # workflow text.
+            res = provider.find(action_name, namespace=namespace)
+
+        if res is None:
+            raise exc.MistralException(
+                "Failed to find action [action_name=%s, namespace=%s]"
+                % (action_name, namespace)
+            )
+
+        return res
 
     def _get_timeout(self):
         timeout = self.task_spec.get_policies().get_timeout()
@@ -649,9 +852,7 @@ class WithItemsTask(RegularTask):
         with_items_values = self._get_with_items_values()
 
         if self._is_new():
-            self._validate_values(with_items_values)
-
-            action_count = len(six.next(iter(with_items_values.values())))
+            action_count = len(next(iter(with_items_values.values())))
 
             self._prepare_runtime_context(action_count)
 
@@ -667,8 +868,6 @@ class WithItemsTask(RegularTask):
 
             action = self._build_action()
 
-            action.validate_input(input_dict)
-
             action.schedule(
                 input_dict,
                 target,
@@ -698,34 +897,37 @@ class WithItemsTask(RegularTask):
 
         :return: Evaluated 'with-items' expression values.
         """
-        ctx_view = data_flow.ContextView(
-            self.ctx,
-            self.wf_ex.context,
-            self.wf_ex.input
-        )
 
-        return expr.evaluate_recursively(
-            self.task_spec.get_with_items(),
-            ctx_view
-        )
+        exp_res = self.evaluate(self.task_spec.get_with_items())
 
-    def _validate_values(self, with_items_values):
-        # Take only mapped values and check them.
-        values = list(with_items_values.values())
+        # Expression result may contain iterables instead of lists in the
+        # dictionary values. So we need to convert them into lists and
+        # perform all needed checks.
 
-        if not all([isinstance(v, list) for v in values]):
-            raise exc.InputException(
-                "Wrong input format for: %s. List type is"
-                " expected for each value." % with_items_values
-            )
+        result = {}
 
-        required_len = len(values[0])
+        required_len = -1
 
-        if not all(len(v) == required_len for v in values):
-            raise exc.InputException(
-                "Wrong input format for: %s. All arrays must"
-                " have the same length." % with_items_values
-            )
+        for var, items in exp_res.items():
+            if not isinstance(items, collections.Iterable):
+                raise exc.InputException(
+                    "Wrong input format for: %s. Iterable type is"
+                    " expected for each value." % result
+                )
+
+            items_list = list(items)
+
+            result[var] = items_list
+
+            if required_len < 0:
+                required_len = len(items_list)
+            elif len(items_list) != required_len:
+                raise exc.InputException(
+                    "Wrong input format for: %s. All arrays must"
+                    " have the same length." % exp_res
+                )
+
+        return result
 
     def _get_input_dicts(self, with_items_values):
         """Calculate input dictionaries for another portion of actions.
@@ -837,10 +1039,11 @@ class WithItemsTask(RegularTask):
             indices = copy.copy(candidates)
 
             if max(candidates) < count - 1:
-                indices += list(six.moves.range(max(candidates) + 1, count))
+                indices += list(range(max(candidates) + 1, count))
         else:
             i = self._get_next_start_index()
-            indices = list(six.moves.range(i, count))
+
+            indices = list(range(i, count))
 
         return indices[:capacity]
 
