@@ -1,6 +1,7 @@
 # Copyright 2016 - Nokia Networks.
 # Copyright 2016 - Brocade Communications Systems, Inc.
 # Copyright 2018 - Extreme Networks, Inc.
+# Copyright 2019 - NetCracker Technology Corp.
 #
 #    Licensed under the Apache License, Version 2.0 (the "License");
 #    you may not use this file except in compliance with the License.
@@ -19,36 +20,33 @@ import json
 from oslo_config import cfg
 from oslo_log import log as logging
 from osprofiler import profiler
-import six
 
 from mistral.db.v2 import api as db_api
 from mistral.db.v2.sqlalchemy import models as db_models
-from mistral.engine import action_queue
 from mistral.engine import dispatcher
+from mistral.engine import post_tx_queue
 from mistral.engine import utils as engine_utils
 from mistral import exceptions as exc
 from mistral import expressions as expr
 from mistral.lang import parser as spec_parser
 from mistral.notifiers import base as notif
 from mistral.notifiers import notification_events as events
+from mistral.rpc import clients as rpc
 from mistral.services import triggers
 from mistral.services import workflows as wf_service
-from mistral import utils
-from mistral.utils import merge_dicts
 from mistral.utils import wf_trace
 from mistral.workflow import base as wf_base
 from mistral.workflow import commands
 from mistral.workflow import data_flow
-from mistral.workflow import lookup_utils
 from mistral.workflow import states
 from mistral_lib import actions as ml_actions
+from mistral_lib import utils
 
 
 LOG = logging.getLogger(__name__)
 
 
-@six.add_metaclass(abc.ABCMeta)
-class Workflow(object):
+class Workflow(object, metaclass=abc.ABCMeta):
     """Workflow.
 
     Represents a workflow and defines interface that can be used by
@@ -66,7 +64,7 @@ class Workflow(object):
         else:
             self.wf_spec = None
 
-    def notify(self, event):
+    def _notify(self, from_state, to_state):
         publishers = self.wf_ex.params.get('notify')
 
         if not publishers and not isinstance(publishers, list):
@@ -74,13 +72,47 @@ class Workflow(object):
 
         notifier = notif.get_notifier(cfg.CONF.notifier.type)
 
-        notifier.notify(
-            self.wf_ex.id,
-            self.wf_ex.to_dict(),
-            event,
-            self.wf_ex.updated_at,
-            publishers
-        )
+        event = events.identify_workflow_event(from_state, to_state)
+
+        filtered_publishers = []
+
+        for publisher in publishers:
+            if not isinstance(publisher, dict):
+                continue
+
+            target_events = publisher.get('event_types', [])
+
+            if not target_events or event in target_events:
+                filtered_publishers.append(publisher)
+
+        if not filtered_publishers:
+            return
+
+        data = {
+            "id": self.wf_ex.id,
+            "name": self.wf_ex.name,
+            "workflow_name": self.wf_ex.workflow_name,
+            "workflow_namespace": self.wf_ex.workflow_namespace,
+            "workflow_id": self.wf_ex.workflow_id,
+            "state": self.wf_ex.state,
+            "state_info": self.wf_ex.state_info,
+            "project_id": self.wf_ex.project_id,
+            "task_execution_id": self.wf_ex.task_execution_id,
+            "root_execution_id": self.wf_ex.root_execution_id,
+            "created_at": utils.datetime_to_str(self.wf_ex.created_at),
+            "updated_at": utils.datetime_to_str(self.wf_ex.updated_at)
+        }
+
+        def _send_notification():
+            notifier.notify(
+                self.wf_ex.id,
+                data,
+                event,
+                self.wf_ex.updated_at,
+                filtered_publishers
+            )
+
+        post_tx_queue.register_operation(_send_notification)
 
     @profiler.trace('workflow-start')
     def start(self, wf_def, wf_ex_id, input_dict, desc='', params=None):
@@ -121,9 +153,6 @@ class Workflow(object):
 
         self.set_state(states.RUNNING)
 
-        # Publish event as soon as state is set to running.
-        self.notify(events.WORKFLOW_LAUNCHED)
-
         wf_ctrl = wf_base.get_controller(self.wf_ex, self.wf_spec)
 
         dispatcher.dispatch_workflow_commands(
@@ -160,9 +189,6 @@ class Workflow(object):
         # Set the state of this workflow to paused.
         self.set_state(states.PAUSED, state_info=msg)
 
-        # Publish event.
-        self.notify(events.WORKFLOW_PAUSED)
-
         # If workflow execution is a subworkflow,
         # schedule update to the task execution.
         if self.wf_ex.task_execution_id:
@@ -182,9 +208,6 @@ class Workflow(object):
 
         self.set_state(states.RUNNING)
 
-        # Publish event.
-        self.notify(events.WORKFLOW_RESUMED)
-
         wf_ctrl = wf_base.get_controller(self.wf_ex)
 
         # Calculate commands to process next.
@@ -215,10 +238,11 @@ class Workflow(object):
             self.wf_spec.__class__.__name__
         )
 
-    def rerun(self, task_ex, reset=True, env=None):
+    def rerun(self, task, reset=True, env=None):
         """Rerun workflow from the given task.
 
-        :param task_ex: Task execution that the workflow needs to rerun from.
+        :param task: An engine task associated with the task the workflow
+            needs to rerun from.
         :param reset: If True, reset task state including deleting its action
             executions.
         :param env: Environment.
@@ -226,28 +250,53 @@ class Workflow(object):
 
         assert self.wf_ex
 
-        # Since some lookup utils functions may use cache for completed tasks
-        # we need to clean caches to make sure that stale objects can't be
-        # retrieved.
-        lookup_utils.clear_caches()
-
         wf_service.update_workflow_execution_env(self.wf_ex, env)
 
-        self.set_state(states.RUNNING, recursive=True)
+        self._recursive_rerun()
 
         wf_ctrl = wf_base.get_controller(self.wf_ex)
 
         # Calculate commands to process next.
-        cmds = wf_ctrl.rerun_tasks([task_ex], reset=reset)
+        cmds = wf_ctrl.rerun_tasks([task.task_ex], reset=reset)
 
         if cmds:
-            # Import the task_handler module here to avoid circular reference.
-            from mistral.engine import policies
-
-            policies.RetryPolicy.refresh_runtime_context(task_ex)
+            task.cleanup_runtime_context()
 
         self._continue_workflow(cmds)
 
+    def _recursive_rerun(self):
+        """Rerun all parent workflow executions recursively.
+
+        If there is a parent execution then it reruns as well.
+        """
+
+        from mistral.engine import workflow_handler
+
+        self.set_state(states.RUNNING)
+
+        # TODO(rakhmerov): We call an internal method of a module here.
+        # The simplest way is to make it public, however, I believe
+        # it's another "bad smell" that tells that some refactoring
+        # of the architecture is needed.
+        workflow_handler._schedule_check_and_fix_integrity(self.wf_ex)
+
+        if self.wf_ex.task_execution_id:
+            parent_task_ex = db_api.get_task_execution(
+                self.wf_ex.task_execution_id
+            )
+
+            parent_wf = Workflow(wf_ex=parent_task_ex.workflow_execution)
+
+            parent_wf.lock()
+
+            parent_wf._recursive_rerun()
+
+            # TODO(rakhmerov): this is a design issue again as in many places.
+            # Ideally, we should just build (or get) an instance of Task and
+            # call set_state() on it.
+            from mistral.engine import task_handler
+            task_handler.mark_task_running(parent_task_ex, parent_wf.wf_spec)
+
     def _get_backlog(self):
         return self.wf_ex.runtime_context.get(dispatcher.BACKLOG_KEY)
 
@@ -297,10 +346,11 @@ class Workflow(object):
         return final_ctx
 
     def _create_execution(self, wf_def, wf_ex_id, input_dict, desc, params):
-        self.wf_ex = db_api.create_workflow_execution({
+        values = {
             'id': wf_ex_id,
             'name': wf_def.name,
             'description': desc,
+            'tags': wf_def.tags,
             'workflow_name': wf_def.name,
             'workflow_namespace': wf_def.namespace,
             'workflow_id': wf_def.id,
@@ -309,10 +359,13 @@ class Workflow(object):
             'output': {},
             'task_execution_id': params.get('task_execution_id'),
             'root_execution_id': params.get('root_execution_id'),
-            'runtime_context': {
-                'index': params.get('index', 0)
-            },
-        })
+            'runtime_context': {'index': params.get('index', 0)}
+        }
+
+        if wf_def.workbook_name:
+            values['runtime_context']['wb_name'] = wf_def.workbook_name
+
+        self.wf_ex = db_api.create_workflow_execution(values)
 
         self.wf_ex.input = input_dict or {}
 
@@ -330,7 +383,7 @@ class Workflow(object):
         )
 
     @profiler.trace('workflow-set-state')
-    def set_state(self, state, state_info=None, recursive=False):
+    def set_state(self, state, state_info=None):
         assert self.wf_ex
 
         cur_state = self.wf_ex.state
@@ -344,12 +397,14 @@ class Workflow(object):
 
             if wf_ex is None:
                 # Do nothing because the state was updated previously.
-                return
+                return False
 
             self.wf_ex = wf_ex
             self.wf_ex.state_info = json.dumps(state_info) \
                 if isinstance(state_info, dict) else state_info
 
+            self._notify(cur_state, state)
+
             wf_trace.info(
                 self.wf_ex,
                 "Workflow '%s' [%s -> %s, msg=%s]" %
@@ -370,27 +425,9 @@ class Workflow(object):
         self.wf_ex.accepted = states.is_completed(state)
 
         if states.is_completed(state):
-            # No need to keep task executions of this workflow in the
-            # lookup cache anymore.
-            lookup_utils.invalidate_cached_task_executions(self.wf_ex.id)
-
             triggers.on_workflow_complete(self.wf_ex)
 
-        if recursive and self.wf_ex.task_execution_id:
-            parent_task_ex = db_api.get_task_execution(
-                self.wf_ex.task_execution_id
-            )
-
-            parent_wf = Workflow(wf_ex=parent_task_ex.workflow_execution)
-
-            parent_wf.lock()
-
-            parent_wf.set_state(state, recursive=recursive)
-
-            # TODO(rakhmerov): It'd be better to use instance of Task here.
-            parent_task_ex.state = state
-            parent_task_ex.state_info = None
-            parent_task_ex.processed = False
+        return True
 
     @profiler.trace('workflow-check-and-complete')
     def check_and_complete(self):
@@ -414,6 +451,16 @@ class Workflow(object):
         if incomplete_tasks_count > 0:
             return incomplete_tasks_count
 
+        LOG.debug("Workflow completed [id=%s]", self.wf_ex.id)
+
+        # NOTE(rakhmerov): Once we know that the workflow has completed,
+        # we need to expire all the objects in the DB session to make sure
+        # to read the most relevant data from the DB (that's already been
+        # committed in parallel transactions). Otherwise, some data like
+        # workflow context may be stale and decisions made upon it will be
+        # wrong.
+        db_api.expire_all()
+
         wf_ctrl = wf_base.get_controller(self.wf_ex, self.wf_spec)
 
         if wf_ctrl.any_cancels():
@@ -433,17 +480,17 @@ class Workflow(object):
         return 0
 
     def _succeed_workflow(self, final_context, msg=None):
-        self.wf_ex.output = data_flow.evaluate_workflow_output(
+        output = data_flow.evaluate_workflow_output(
             self.wf_ex,
             self.wf_spec.get_output(),
             final_context
         )
 
-        # Set workflow execution to success until after output is evaluated.
-        self.set_state(states.SUCCESS, msg)
+        # Set workflow execution to success after output is evaluated.
+        if not self.set_state(states.SUCCESS, msg):
+            return
 
-        # Publish event.
-        self.notify(events.WORKFLOW_SUCCEEDED)
+        self.wf_ex.output = output
 
         if self.wf_ex.task_execution_id:
             self._send_result_to_parent_workflow()
@@ -468,7 +515,8 @@ class Workflow(object):
             )
             LOG.error(msg)
 
-        self.set_state(states.ERROR, state_info=msg)
+        if not self.set_state(states.ERROR, state_info=msg):
+            return
 
         # When we set an ERROR state we should safely set output value getting
         # w/o exceptions due to field size limitations.
@@ -488,10 +536,7 @@ class Workflow(object):
                 cfg.CONF.engine.execution_field_size_limit_kb
             )
 
-        self.wf_ex.output = merge_dicts({'result': msg}, output_on_error)
-
-        # Publish event.
-        self.notify(events.WORKFLOW_FAILED)
+        self.wf_ex.output = utils.merge_dicts({'result': msg}, output_on_error)
 
         if self.wf_ex.task_execution_id:
             self._send_result_to_parent_workflow()
@@ -500,7 +545,8 @@ class Workflow(object):
         if states.is_completed(self.wf_ex.state):
             return
 
-        self.set_state(states.CANCELLED, state_info=msg)
+        if not self.set_state(states.CANCELLED, state_info=msg):
+            return
 
         # When we set an ERROR state we should safely set output value getting
         # w/o exceptions due to field size limitations.
@@ -511,15 +557,14 @@ class Workflow(object):
 
         self.wf_ex.output = {'result': msg}
 
-        # Publish event.
-        self.notify(events.WORKFLOW_CANCELLED)
-
         if self.wf_ex.task_execution_id:
             self._send_result_to_parent_workflow()
 
     def _send_result_to_parent_workflow(self):
         if self.wf_ex.state == states.SUCCESS:
-            result = ml_actions.Result(data=self.wf_ex.output)
+            # The result of the sub workflow is already saved
+            # so there's no need to send it over RPC.
+            result = None
         elif self.wf_ex.state == states.ERROR:
             err_msg = (
                 self.wf_ex.state_info or
@@ -540,11 +585,16 @@ class Workflow(object):
                 " if a workflow is not in SUCCESS, ERROR or CANCELLED state."
             )
 
-        action_queue.schedule_on_action_complete(
-            self.wf_ex.id,
-            result,
-            wf_action=True
-        )
+        # Register a command executed in a separate thread to send the result
+        # to the parent workflow outside of the main DB transaction.
+        def _send_result():
+            rpc.get_engine_client().on_action_complete(
+                self.wf_ex.id,
+                result,
+                wf_action=True
+            )
+
+        post_tx_queue.register_operation(_send_result)
 
 
 def _get_environment(params):
@@ -555,7 +605,7 @@ def _get_environment(params):
 
     if isinstance(env, dict):
         env_dict = env
-    elif isinstance(env, six.string_types):
+    elif isinstance(env, str):
         env_db = db_api.load_environment(env)
 
         if not env_db:
@@ -579,13 +629,13 @@ def _get_environment(params):
 
 def _build_fail_info_message(wf_ctrl, wf_ex):
     # Try to find where error is exactly.
-    failed_tasks = sorted(
-        filter(
-            lambda t: not wf_ctrl.is_error_handled_for(t),
-            lookup_utils.find_error_task_executions(wf_ex.id)
-        ),
-        key=lambda t: t.name
-    )
+    failed_tasks = [
+        t_ex for t_ex in db_api.get_task_executions(
+            workflow_execution_id=wf_ex.id,
+            state=states.ERROR,
+            sort_keys=['name']
+        ) if not wf_ctrl.is_error_handled_for(t_ex)
+    ]
 
     msg = ('Failure caused by error in tasks: %s\n' %
            ', '.join([t.name for t in failed_tasks]))
@@ -620,10 +670,13 @@ def _build_fail_info_message(wf_ctrl, wf_ex):
 
 def _build_cancel_info_message(wf_ctrl, wf_ex):
     # Try to find where cancel is exactly.
-    cancelled_tasks = sorted(
-        lookup_utils.find_cancelled_task_executions(wf_ex.id),
-        key=lambda t: t.name
-    )
+    cancelled_tasks = [
+        t_ex for t_ex in db_api.get_task_executions(
+            workflow_execution_id=wf_ex.id,
+            state=states.CANCELLED,
+            sort_keys=['name']
+        )
+    ]
 
     return (
         'Cancelled tasks: %s' % ', '.join([t.name for t in cancelled_tasks])
