@@ -15,17 +15,16 @@
 import copy
 import datetime
 import json
-import uuid
+from unittest import mock
 
-import mock
-import six
+import sqlalchemy as sa
 
 from mistral.api.controllers.v2 import resources
 from mistral.db.v2 import api as db_api
 from mistral.db.v2.sqlalchemy import models as db
 from mistral import exceptions as exc
 from mistral.tests.unit.api import base
-
+from oslo_utils import uuidutils
 
 DATETIME_FORMAT = '%Y-%m-%d %H:%M:%S.%f'
 
@@ -36,7 +35,7 @@ VARIABLES = {
     'verbose': True,
     '__actions': {
         'std.sql': {
-            'conn': 'mysql://admin:secrete@<% env().host %>/<% env().db %>'
+            'conn': 'mysql://admin:secret@<% env().host %>/<% env().db %>'
         }
     }
 }
@@ -62,17 +61,18 @@ ENVIRONMENT_FOR_UPDATE_NO_SCOPE = {
 
 
 ENVIRONMENT = {
-    'id': str(uuid.uuid4()),
+    'id': uuidutils.generate_uuid(),
     'name': 'test',
     'description': 'my test settings',
     'variables': VARIABLES,
     'scope': 'private',
+    'project_id': '<default-project>',
     'created_at': str(datetime.datetime.utcnow()),
     'updated_at': str(datetime.datetime.utcnow())
 }
 
 ENVIRONMENT_WITH_ILLEGAL_FIELD = {
-    'id': str(uuid.uuid4()),
+    'id': uuidutils.generate_uuid(),
     'name': 'test',
     'description': 'my test settings',
     'extra_field': 'I can add whatever I want here',
@@ -86,13 +86,17 @@ ENVIRONMENT_DB = db.Environment(
     description=ENVIRONMENT['description'],
     variables=copy.deepcopy(VARIABLES),
     scope=ENVIRONMENT['scope'],
+    project_id=ENVIRONMENT['project_id'],
     created_at=datetime.datetime.strptime(ENVIRONMENT['created_at'],
                                           DATETIME_FORMAT),
     updated_at=datetime.datetime.strptime(ENVIRONMENT['updated_at'],
                                           DATETIME_FORMAT)
 )
 
-ENVIRONMENT_DB_DICT = {k: v for k, v in six.iteritems(ENVIRONMENT_DB)}
+ENVIRONMENT_DB_WITH_PROJECT_ID = ENVIRONMENT_DB.get_clone()
+ENVIRONMENT_DB_WITH_PROJECT_ID.project_id = '<default-project>'
+
+ENVIRONMENT_DB_DICT = {k: v for k, v in ENVIRONMENT_DB.items()}
 
 UPDATED_VARIABLES = copy.deepcopy(VARIABLES)
 UPDATED_VARIABLES['host'] = '127.0.0.1'
@@ -115,7 +119,7 @@ MOCK_DELETE = mock.MagicMock(return_value=None)
 def _convert_vars_to_dict(env_dict):
     """Converts 'variables' in the given environment dict into dictionary."""
     if ('variables' in env_dict and
-            isinstance(env_dict.get('variables'), six.string_types)):
+            isinstance(env_dict.get('variables'), str)):
         env_dict['variables'] = json.loads(env_dict['variables'])
 
     return env_dict
@@ -156,6 +160,21 @@ class TestEnvironmentController(base.APITest):
         self.assertEqual(200, resp.status_int)
         self.assertEqual(1, len(resp.json['environments']))
 
+    @mock.patch.object(db_api, 'get_environments')
+    def test_get_all_operational_error(self, mocked_get_all):
+        mocked_get_all.side_effect = [
+            # Emulating DB OperationalError
+            sa.exc.OperationalError('Mock', 'mock', 'mock'),
+            [ENVIRONMENT_DB]  # Successful run
+        ]
+
+        resp = self.app.get('/v2/environments')
+
+        self.assertEqual(200, resp.status_int)
+
+        self.assertEqual(1, len(resp.json['environments']))
+        self._assert_dict_equal(ENVIRONMENT, resp.json['environments'][0])
+
     def test_get_all_empty(self):
         resp = self.app.get('/v2/environments')
 
@@ -169,6 +188,27 @@ class TestEnvironmentController(base.APITest):
         self.assertEqual(200, resp.status_int)
         self._assert_dict_equal(ENVIRONMENT, resp.json)
 
+    @mock.patch.object(db_api, 'get_environment')
+    def test_get_operational_error(self, mocked_get):
+        mocked_get.side_effect = [
+            # Emulating DB OperationalError
+            sa.exc.OperationalError('Mock', 'mock', 'mock'),
+            ENVIRONMENT_DB  # Successful run
+        ]
+
+        resp = self.app.get('/v2/environments/123')
+
+        self.assertEqual(200, resp.status_int)
+        self._assert_dict_equal(ENVIRONMENT, resp.json)
+
+    @mock.patch.object(db_api, 'get_environment',
+                       return_value=ENVIRONMENT_DB_WITH_PROJECT_ID)
+    def test_get_within_project_id(self, mock_get):
+        resp = self.app.get('/v2/environments/123')
+
+        self.assertEqual(200, resp.status_int)
+        self.assertEqual('<default-project>', resp.json['project_id'])
+
     @mock.patch.object(db_api, "get_environment", MOCK_NOT_FOUND)
     def test_get_not_found(self):
         resp = self.app.get('/v2/environments/123', expect_errors=True)
