@@ -19,8 +19,8 @@
 import copy
 import datetime
 import json
+from unittest import mock
 
-import mock
 from oslo_config import cfg
 import oslo_messaging
 from oslo_utils import uuidutils
@@ -28,6 +28,7 @@ import sqlalchemy as sa
 from webtest import app as webtest_app
 
 from mistral.api.controllers.v2 import execution
+from mistral.api.controllers.v2 import resources
 from mistral.db.v2 import api as db_api
 from mistral.db.v2.sqlalchemy import api as sql_db_api
 from mistral.db.v2.sqlalchemy import models
@@ -36,9 +37,9 @@ from mistral.rpc import base as rpc_base
 from mistral.rpc import clients as rpc_clients
 from mistral.tests.unit.api import base
 from mistral.tests.unit import base as unit_base
-from mistral import utils
 from mistral.utils import rest_utils
 from mistral.workflow import states
+from mistral_lib import utils
 
 # This line is needed for correct initialization of messaging config.
 oslo_messaging.get_rpc_transport(cfg.CONF)
@@ -52,6 +53,7 @@ WF_EX = models.WorkflowExecution(
     spec={'name': 'some'},
     state=states.RUNNING,
     state_info=None,
+    context={},
     input={'foo': 'bar'},
     output={},
     params={'env': {'k1': 'abc'}},
@@ -80,6 +82,7 @@ SUB_WF_EX = models.WorkflowExecution(
     spec={'name': 'some'},
     state=states.RUNNING,
     state_info=None,
+    context={},
     input={'foo': 'bar'},
     output={},
     params={'env': {'k1': 'abc'}},
@@ -160,7 +163,11 @@ class TestExecutionsController(base.APITest):
         resp = self.app.get('/v2/executions/123')
 
         self.assertEqual(200, resp.status_int)
-        self.assertDictEqual(WF_EX_JSON_WITH_DESC, resp.json)
+
+        expected = WF_EX_JSON_WITH_DESC.copy()
+        expected['published_global'] = '{}'
+
+        self.assertDictEqual(expected, resp.json)
 
     @mock.patch.object(db_api, 'get_workflow_execution')
     def test_get_operational_error(self, mocked_get):
@@ -173,14 +180,22 @@ class TestExecutionsController(base.APITest):
         resp = self.app.get('/v2/executions/123')
 
         self.assertEqual(200, resp.status_int)
-        self.assertDictEqual(WF_EX_JSON_WITH_DESC, resp.json)
+
+        expected = WF_EX_JSON_WITH_DESC.copy()
+        expected['published_global'] = '{}'
+
+        self.assertDictEqual(expected, resp.json)
 
     @mock.patch.object(db_api, 'get_workflow_execution', MOCK_SUB_WF_EX)
     def test_get_sub_wf_ex(self):
         resp = self.app.get('/v2/executions/123')
 
         self.assertEqual(200, resp.status_int)
-        self.assertDictEqual(SUB_WF_EX_JSON_WITH_DESC, resp.json)
+
+        expected = SUB_WF_EX_JSON_WITH_DESC.copy()
+        expected['published_global'] = '{}'
+
+        self.assertDictEqual(expected, resp.json)
 
     @mock.patch.object(db_api, 'get_workflow_execution', MOCK_NOT_FOUND)
     def test_get_not_found(self):
@@ -192,6 +207,7 @@ class TestExecutionsController(base.APITest):
                        return_value=WF_EX_WITH_PROJECT_ID)
     def test_get_within_project_id(self, mock_get):
         resp = self.app.get('/v2/executions/123', expect_errors=True)
+
         self.assertEqual(200, resp.status_int)
         self.assertTrue('project_id' in resp.json)
 
@@ -683,6 +699,23 @@ class TestExecutionsController(base.APITest):
             **json.loads(exec_dict['params'])
         )
 
+    @mock.patch.object(rpc_clients.EngineClient, 'start_workflow')
+    def test_post_with_params_none(self, start_wf_func):
+        wf_ex_dict = WF_EX.to_dict()
+
+        start_wf_func.return_value = wf_ex_dict
+
+        json_body = WF_EX_JSON_WITH_DESC.copy()
+
+        json_body['params'] = None
+
+        expected_json = WF_EX_JSON_WITH_DESC
+
+        resp = self.app.post_json('/v2/executions', json_body)
+
+        self.assertEqual(201, resp.status_int)
+        self.assertDictEqual(expected_json, resp.json)
+
     @mock.patch.object(
         rpc_clients.EngineClient,
         'start_workflow',
@@ -844,7 +877,7 @@ class TestExecutionsController(base.APITest):
 
     def test_get_all_pagination_unknown_direction(self):
         resp = self.app.get(
-            '/v2/actions?limit=1&sort_keys=id&sort_dirs=nonexist',
+            '/v2/executions?limit=1&sort_keys=id&sort_dirs=nonexist',
             expect_errors=True
         )
 
@@ -871,7 +904,8 @@ class TestExecutionsController(base.APITest):
         )
 
     @mock.patch.object(db_api, 'get_workflow_executions', MOCK_WF_EXECUTIONS)
-    @mock.patch.object(rest_utils, 'get_all')
+    @mock.patch.object(rest_utils, 'get_all',
+                       return_value=resources.Executions())
     def test_get_all_executions_with_output(self, mock_get_all):
         resp = self.app.get('/v2/executions?include_output=true')
 
@@ -880,23 +914,27 @@ class TestExecutionsController(base.APITest):
         args, kwargs = mock_get_all.call_args
         resource_function = kwargs['resource_function']
 
+        self.assertEqual(
+            execution._get_workflow_execution_resource_with_output,
+            resource_function
+        )
+
+    @mock.patch.object(db_api, 'get_workflow_executions', MOCK_WF_EXECUTIONS)
+    @mock.patch.object(rest_utils, 'get_all',
+                       return_value=resources.Executions())
+    def test_get_all_executions_without_output(self, mock_get_all):
+        resp = self.app.get('/v2/executions')
+
+        self.assertEqual(200, resp.status_int)
+
+        args, kwargs = mock_get_all.call_args
+        resource_function = kwargs['resource_function']
+
         self.assertEqual(
             execution._get_workflow_execution_resource,
             resource_function
         )
 
-    @mock.patch.object(db_api, 'get_workflow_executions', MOCK_WF_EXECUTIONS)
-    @mock.patch.object(rest_utils, 'get_all')
-    def test_get_all_executions_without_output(self, mock_get_all):
-        resp = self.app.get('/v2/executions')
-
-        self.assertEqual(200, resp.status_int)
-
-        args, kwargs = mock_get_all.call_args
-        resource_function = kwargs['resource_function']
-
-        self.assertIsNone(resource_function)
-
     @mock.patch('mistral.db.v2.api.get_workflow_executions')
     @mock.patch('mistral.context.MistralContext.from_environ')
     def test_get_all_projects_admin(self, mock_context, mock_get_execs):
@@ -933,3 +971,26 @@ class TestExecutionsController(base.APITest):
         self.assertTrue(
             mock_get_execs.call_args[1].get('project_id', fake_project_id)
         )
+
+    def test_get_all_with_nulls_not_valid(self):
+        resp = self.app.get(
+            '/v2/executions?limit=10&sort_keys=id&sort_dirs=asc&nulls=invalid',
+            expect_errors=True
+        )
+
+        self.assertEqual(500, resp.status_int)
+        self.assertIn(
+            "'invalid' is not a valid field name.",
+            resp.body.decode()
+        )
+
+        resp = self.app.get(
+            '/v2/executions?limit=10&sort_keys=id&sort_dirs=asc&nulls=id',
+            expect_errors=True
+        )
+
+        self.assertEqual(500, resp.status_int)
+        self.assertIn(
+            "The field 'id' can't hold None value.",
+            resp.body.decode()
+        )
