@@ -1,5 +1,6 @@
 # Copyright 2014 - Mirantis, Inc.
 # Copyright 2015 - StackStorm, Inc.
+# Copyright 2020 Nokia Software.
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -18,6 +19,7 @@ from oslo_config import cfg
 from mistral.db.v2 import api as db_api
 from mistral import exceptions as exc
 from mistral.lang import parser as spec_parser
+from mistral.scheduler import base as sched_base
 from mistral.services import workflows as wf_service
 from mistral.tests.unit.engine import base
 from mistral.workflow import states
@@ -88,7 +90,7 @@ class DirectWorkflowEngineTest(base.EngineTestCase):
         self.await_task_success(task1.id)
         self.await_task_success(task2.id)
 
-        self.assertTrue(wf_ex.state, states.ERROR)
+        self.assertEqual(wf_ex.state, states.SUCCESS)
 
     def test_condition_transition_not_triggering(self):
         wf_text = """---
@@ -129,7 +131,7 @@ class DirectWorkflowEngineTest(base.EngineTestCase):
 
         self.await_task_error(task1.id)
 
-        self.assertTrue(wf_ex.state, states.ERROR)
+        self.assertEqual(wf_ex.state, states.ERROR)
 
     def test_change_state_after_success(self):
         wf_text = """
@@ -221,7 +223,7 @@ class DirectWorkflowEngineTest(base.EngineTestCase):
         wf_ex = self._run_workflow(wf_text)
 
         self.assertIn('Invalid input', wf_ex.state_info)
-        self.assertTrue(wf_ex.state, states.ERROR)
+        self.assertEqual(wf_ex.state, states.ERROR)
 
     def test_wrong_first_task_input(self):
         wf_text = """
@@ -262,7 +264,7 @@ class DirectWorkflowEngineTest(base.EngineTestCase):
         # TODO(dzimine): Catch tasks caused error, and set them to ERROR:
         # TODO(dzimine): self.assertTrue(task_ex.state, states.ERROR)
 
-        self.assertTrue(wf_ex.state, states.ERROR)
+        self.assertEqual(wf_ex.state, states.ERROR)
         self.assertIn("Failed to find action", wf_ex.state_info)
 
     def test_wrong_action_first_task(self):
@@ -282,7 +284,7 @@ class DirectWorkflowEngineTest(base.EngineTestCase):
         wf_ex = self.engine.start_workflow('wf')
 
         self.assertIn(
-            "Failed to find action [action_name=wrong.task]",
+            "Failed to find action [action_name=wrong.task, namespace=]",
             wf_ex.state_info
         )
         self.assertEqual(states.ERROR, wf_ex.state)
@@ -737,7 +739,7 @@ class DirectWorkflowEngineTest(base.EngineTestCase):
 
         self.assertIn("Task 'task3' not found", str(exception))
 
-    def test_delete_workflow_completion_check_on_stop(self):
+    def test_delete_workflow_integrity_check_on_stop(self):
         wf_text = """---
         version: '2.0'
 
@@ -751,20 +753,13 @@ class DirectWorkflowEngineTest(base.EngineTestCase):
 
         wf_ex = self.engine.start_workflow('wf')
 
-        calls = db_api.get_delayed_calls()
-
-        mtd_name = 'mistral.engine.workflow_handler._check_and_complete'
-
-        self._assert_single_item(calls, target_method_name=mtd_name)
-
         self.engine.stop_workflow(wf_ex.id, state=states.CANCELLED)
 
-        self._await(
-            lambda:
-            len(db_api.get_delayed_calls(target_method_name=mtd_name)) == 0
-        )
+        sched = sched_base.get_system_scheduler()
 
-    def test_delete_workflow_completion_on_execution_delete(self):
+        self._await(lambda: not sched.has_scheduled_jobs())
+
+    def test_delete_workflow_integrity_check_on_execution_delete(self):
         wf_text = """---
         version: '2.0'
 
@@ -778,18 +773,11 @@ class DirectWorkflowEngineTest(base.EngineTestCase):
 
         wf_ex = self.engine.start_workflow('wf')
 
-        calls = db_api.get_delayed_calls()
-
-        mtd_name = 'mistral.engine.workflow_handler._check_and_complete'
-
-        self._assert_single_item(calls, target_method_name=mtd_name)
-
         db_api.delete_workflow_execution(wf_ex.id)
 
-        self._await(
-            lambda:
-            len(db_api.get_delayed_calls(target_method_name=mtd_name)) == 0
-        )
+        sched = sched_base.get_system_scheduler()
+
+        self._await(lambda: not sched.has_scheduled_jobs())
 
     def test_output(self):
         wf_text = """---
@@ -816,6 +804,44 @@ class DirectWorkflowEngineTest(base.EngineTestCase):
 
             self.assertDictEqual({}, wf_ex.output)
 
+    def test_output_expression(self):
+        wf_text = """---
+        version: '2.0'
+
+        wf:
+          output:
+            continue_flag: <% $.continue_flag %>
+
+          task-defaults:
+            on-error:
+              - task2
+
+          tasks:
+            task1:
+              action: std.fail
+              on-success: task3
+
+            task2:
+              action: std.noop
+              publish:
+                continue_flag: false
+
+            task3:
+              action: std.noop
+        """
+
+        wf_service.create_workflows(wf_text)
+
+        wf_ex = self.engine.start_workflow('wf')
+
+        self.await_workflow_success(wf_ex.id)
+
+        with db_api.transaction():
+            wf_ex = db_api.get_workflow_execution(wf_ex.id)
+
+            self.assertEqual(2, len(wf_ex.task_executions))
+            self.assertDictEqual({'continue_flag': False}, wf_ex.output)
+
     def test_triggered_by(self):
         wf_text = """---
         version: '2.0'
@@ -1006,7 +1032,7 @@ class DirectWorkflowEngineTest(base.EngineTestCase):
         """
 
         # Generate the workflow text.
-        task_cnt = 200
+        task_cnt = 50
 
         on_success_list_str = ''
 
@@ -1052,3 +1078,113 @@ class DirectWorkflowEngineTest(base.EngineTestCase):
 
         self._assert_single_item(task_execs, name='task0')
         self._assert_single_item(task_execs, name='task{}'.format(task_cnt))
+
+    def test_action_error_with_array_result(self):
+        wf_text = """---
+        version: '2.0'
+
+        wf:
+          tasks:
+            task1:
+              action: std.fail error_data=[1,2,3]
+        """
+
+        wf_service.create_workflows(wf_text)
+
+        # Start workflow.
+        wf_ex = self.engine.start_workflow('wf')
+
+        self.await_workflow_error(wf_ex.id)
+
+        with db_api.transaction():
+            # Note: We need to reread execution to access related tasks.
+            wf_ex = db_api.get_workflow_execution(wf_ex.id)
+
+            task_ex = self._assert_single_item(
+                wf_ex.task_executions,
+                name='task1',
+                state=states.ERROR
+            )
+
+            a_ex = self._assert_single_item(
+                task_ex.action_executions,
+                name='std.fail'
+            )
+
+            self.assertIsInstance(a_ex.output.get('result'), list)
+
+            # NOTE(rakhmerov): This was previously failing but only Python 2.7
+            # probably because SQLAlchemy works differently on different
+            # versions of Python. On Python 3 this field's value was always
+            # converted into a string no matter what we tried to assign. But
+            # that didn't happen on Python 2.7 which caused an SQL exception.
+            self.assertIsInstance(task_ex.state_info, str)
+
+    def test_single_fail_with_next_noop(self):
+        wf_text = """---
+        version: '2.0'
+
+        wf:
+          tasks:
+            task1:
+              action: std.fail
+              on-error:
+                - noop
+        """
+
+        wf_service.create_workflows(wf_text)
+
+        # Start workflow.
+        wf_ex = self.engine.start_workflow('wf')
+
+        self.await_workflow_success(wf_ex.id)
+
+    def test_unexisting_join_task_does_not_stuck_wf_running(self):
+        wf_text = """---
+        version: '2.0'
+
+        wf:
+          tasks:
+            branch1:
+              action: std.noop
+              on-success: branch1-23_merge
+            branch2:
+              action: std.async_noop
+              on-success: branch2-3_merge
+            branch3:
+              action: std.fail
+              on-success: branch2-3_merge
+            branch2-3_merge:
+              action: std.noop
+              on-success: branch1-23_merge
+              join: all
+            branch1-23_merge:
+              action: std.noop
+              join: all
+        """
+
+        wf_service.create_workflows(wf_text)
+
+        # Start workflow.
+        wf_ex = self.engine.start_workflow('wf')
+
+        with db_api.transaction():
+            wf_ex = db_api.get_workflow_execution(wf_ex.id)
+
+            task_execs = wf_ex.task_executions
+
+        t_ex = self._assert_single_item(
+            task_execs,
+            name='branch2'
+        )
+
+        t_action_exs = db_api.get_action_executions(
+            task_execution_id=t_ex.id
+        )
+
+        self.engine.on_action_complete(
+            t_action_exs[0].id,
+            ml_actions.Result(error="Error!")
+        )
+
+        self.await_workflow_error(wf_ex.id)
