@@ -14,15 +14,13 @@
 
 from oslo_config import cfg
 
-from mistral.actions import base as actions_base
+from mistral_lib import actions as actions_base
+
 from mistral.db.v2 import api as db_api
 from mistral import exceptions as exc
 from mistral.services import workflows as wf_service
-from mistral.tests.unit import base as test_base
 from mistral.tests.unit.engine import base
 from mistral.workflow import states
-from mistral.workflow import utils as wf_utils
-
 
 # Use the set_default method to set value otherwise in certain test cases
 # the change in value is not permanent.
@@ -36,27 +34,42 @@ wf:
   input:
     - workflow_input: '__WORKFLOW_INPUT__'
     - action_output_length: 0
+    - action_output_dict: false
+    - action_error: false
 
   tasks:
     task1:
       action: my_action
       input:
-        action_input: '__ACTION_INPUT__'
-        action_output_length: <% $.action_output_length %>
+        input: '__ACTION_INPUT__'
+        output_length: <% $.action_output_length %>
+        output_dict: <% $.action_output_dict %>
+        error: <% $.action_error %>
       publish:
         p_var: '__TASK_PUBLISHED__'
 """
 
 
 class MyAction(actions_base.Action):
-    def __init__(self, action_input, action_output_length):
-        self.action_input = action_input
-        self.action_output_length = action_output_length
+    def __init__(self, input, output_length, output_dict=False, error=False):
+        self.input = input
+        self.output_length = output_length
+        self.output_dict = output_dict
+        self.error = error
 
-    def run(self):
-        return wf_utils.Result(
-            data=''.join('A' for _ in range(self.action_output_length))
-        )
+    def run(self, context):
+        if not self.output_dict:
+            result = ''.join('A' for _ in range(self.output_length))
+        else:
+            result = {}
+
+            for i in range(self.output_length):
+                result[i] = 'A'
+
+        if not self.error:
+            return actions_base.Result(data=result)
+        else:
+            return actions_base.Result(error=result)
 
     def test(self):
         raise NotImplementedError
@@ -83,7 +96,7 @@ class ExecutionFieldsSizeLimitTest(base.EngineTestCase):
             group='engine'
         )
 
-        test_base.register_action_class('my_action', MyAction)
+        self.register_action_class('my_action', MyAction)
 
     def tearDown(self):
         """Restores the size limit config to default"""
@@ -109,7 +122,7 @@ class ExecutionFieldsSizeLimitTest(base.EngineTestCase):
         wf_service.create_workflows(new_wf)
 
         # Start workflow.
-        wf_ex = self.engine.start_workflow('wf', {})
+        wf_ex = self.engine.start_workflow('wf')
 
         self.await_workflow_success(wf_ex.id)
 
@@ -122,13 +135,13 @@ class ExecutionFieldsSizeLimitTest(base.EngineTestCase):
         e = self.assertRaises(
             exc.SizeLimitExceededException,
             self.engine.start_workflow,
-            'wf',
-            {}
+            'wf'
         )
 
         self.assertEqual(
-            "Size of 'input' is 1KB which exceeds the limit of 0KB",
-            e.message
+            'Field size limit exceeded'
+            ' [class=TaskExecution, field=input, size=1KB, limit=0KB]',
+            str(e)
         )
 
     def test_workflow_input_limit(self):
@@ -139,12 +152,13 @@ class ExecutionFieldsSizeLimitTest(base.EngineTestCase):
             exc.SizeLimitExceededException,
             self.engine.start_workflow,
             'wf',
-            {'workflow_input': ''.join('A' for _ in range(1024))}
+            wf_input={'workflow_input': ''.join('A' for _ in range(1024))}
         )
 
         self.assertEqual(
-            "Size of 'input' is 1KB which exceeds the limit of 0KB",
-            e.message
+            'Field size limit exceeded'
+            ' [class=TaskExecution, field=input, size=1KB, limit=0KB]',
+            str(e)
         )
 
     def test_action_input_limit(self):
@@ -153,11 +167,12 @@ class ExecutionFieldsSizeLimitTest(base.EngineTestCase):
         wf_service.create_workflows(new_wf)
 
         # Start workflow.
-        wf_ex = self.engine.start_workflow('wf', {})
+        wf_ex = self.engine.start_workflow('wf')
 
         self.assertEqual(states.ERROR, wf_ex.state)
         self.assertIn(
-            "Size of 'input' is 1KB which exceeds the limit of 0KB",
+            "Field size limit exceeded"
+            " [class=TaskExecution, field=input, size=1KB, limit=0KB]",
             wf_ex.state_info
         )
 
@@ -167,7 +182,7 @@ class ExecutionFieldsSizeLimitTest(base.EngineTestCase):
         # Start workflow.
         wf_ex = self.engine.start_workflow(
             'wf',
-            {'action_output_length': 1024}
+            wf_input={'action_output_length': 1024}
         )
 
         self.await_workflow_error(wf_ex.id)
@@ -176,7 +191,8 @@ class ExecutionFieldsSizeLimitTest(base.EngineTestCase):
         wf_ex = db_api.get_workflow_execution(wf_ex.id)
 
         self.assertIn(
-            "Size of 'output' is 1KB which exceeds the limit of 0KB",
+            'Field size limit exceeded'
+            ' [class=TaskExecution, field=output, size=1KB, limit=0KB]',
             wf_ex.state_info
         )
         self.assertEqual(states.ERROR, wf_ex.state)
@@ -187,22 +203,27 @@ class ExecutionFieldsSizeLimitTest(base.EngineTestCase):
         wf_service.create_workflows(new_wf)
 
         # Start workflow.
-        wf_ex = self.engine.start_workflow('wf', {})
+        wf_ex = self.engine.start_workflow('wf')
 
         self.await_workflow_error(wf_ex.id)
 
-        # Note: We need to reread execution to access related tasks.
-        wf_ex = db_api.get_workflow_execution(wf_ex.id)
+        with db_api.transaction():
+            # Note: We need to reread execution to access related tasks.
+            wf_ex = db_api.get_workflow_execution(wf_ex.id)
+
+            task_execs = wf_ex.task_executions
 
         self.assertIn(
-            'Failed to handle action completion [wf=wf, task=task1',
+            'Failed to handle action completion [error=Field size',
             wf_ex.state_info
         )
+        self.assertIn('wf=wf, task=task1', wf_ex.state_info)
 
-        task_ex = self._assert_single_item(wf_ex.task_executions, name='task1')
+        task_ex = self._assert_single_item(task_execs, name='task1')
 
         self.assertIn(
-            "Size of 'published' is 1KB which exceeds the limit of 0KB",
+            'Field size limit exceeded'
+            ' [class=TaskExecution, field=published, size=1KB, limit=0KB]',
             task_ex.state_info
         )
 
@@ -216,12 +237,73 @@ class ExecutionFieldsSizeLimitTest(base.EngineTestCase):
             exc.SizeLimitExceededException,
             self.engine.start_workflow,
             'wf',
-            {},
-            '',
             env={'param': long_string}
         )
 
         self.assertIn(
-            "Size of 'params' is 1KB which exceeds the limit of 0KB",
-            e.message
+            'Field size limit exceeded'
+            ' [class=TaskExecution, field=params, size=1KB, limit=0KB]',
+            str(e)
         )
+
+    def test_task_execution_state_info_trimmed(self):
+        # No limit on output, input and other JSON fields.
+        cfg.CONF.set_default(
+            'execution_field_size_limit_kb',
+            -1,
+            group='engine'
+        )
+
+        wf_service.create_workflows(WF)
+
+        # Start workflow.
+        wf_ex = self.engine.start_workflow(
+            'wf',
+            wf_input={
+                'action_output_length': 80000,
+                'action_output_dict': True,
+                'action_error': True
+            }
+        )
+
+        self.await_workflow_error(wf_ex.id)
+
+        with db_api.transaction():
+            wf_ex = db_api.get_workflow_execution(wf_ex.id)
+
+            task_ex = self._assert_single_item(
+                wf_ex.task_executions,
+                state=states.ERROR
+            )
+
+            # "state_info" must be trimmed so that it's not greater than 65535.
+            self.assertLess(len(task_ex.state_info), 65536)
+            self.assertGreater(len(task_ex.state_info), 65490)
+            self.assertLess(len(wf_ex.state_info), 65536)
+            self.assertGreater(len(wf_ex.state_info), 65490)
+
+    def test_fail_workflow_no_limit(self):
+        cfg.CONF.set_default(
+            'execution_field_size_limit_kb',
+            -1,
+            group='engine'
+        )
+
+        wf_service.create_workflows(WF)
+
+        # Start workflow.
+        wf_ex = self.engine.start_workflow(
+            'wf',
+            wf_input={
+                'action_output_length': 10000,
+                'action_output_dict': True,
+                'action_error': True
+            }
+        )
+
+        self.await_workflow_error(wf_ex.id)
+
+        with db_api.transaction():
+            wf_ex = db_api.get_workflow_execution(wf_ex.id)
+
+            self.assertGreater(len(wf_ex.output['result']), 10000)
