@@ -13,10 +13,11 @@
 #    limitations under the License.
 
 import json
-import mock
+from unittest import mock
 
 from oslo_config import cfg
 
+from mistral import context
 from mistral.db.v2 import api as db_api
 from mistral.notifiers import base as notif
 from mistral.notifiers import notification_events as events
@@ -33,12 +34,14 @@ cfg.CONF.set_default('auth_enable', False, group='pecan')
 EVENT_LOGS = []
 
 
-def log_event(ex_id, data, event, timestamp, **kwargs):
+def log_event(ctx, ex_id, data, event, timestamp, **kwargs):
+    if not isinstance(ctx, context.MistralContext):
+        raise TypeError('ctx is not type of MistralContext.')
+
     EVENT_LOGS.append((ex_id, event))
 
 
 class NotifyEventsTest(base.NotifierTestCase):
-
     def setUp(self):
         super(NotifyEventsTest, self).setUp()
 
@@ -53,15 +56,18 @@ class NotifyEventsTest(base.NotifierTestCase):
         self.publishers['noop'].publish.reset_mock()
 
         del EVENT_LOGS[:]
+
         cfg.CONF.set_default('type', 'local', group='notifier')
 
     def tearDown(self):
-        cfg.CONF.set_default('notify', None, group='notifier')
         super(NotifyEventsTest, self).tearDown()
 
+        cfg.CONF.set_default('notify', None, group='notifier')
+
     def test_notify_all_explicit(self):
-        wf_def = """
+        wf_text = """
         version: '2.0'
+
         wf:
           tasks:
             t1:
@@ -72,12 +78,12 @@ class NotifyEventsTest(base.NotifierTestCase):
               action: std.noop
         """
 
-        wf_svc.create_workflows(wf_def)
+        wf_svc.create_workflows(wf_text)
 
         notify_options = [
             {
                 'type': 'webhook',
-                'events': events.EVENTS
+                'event_types': events.EVENTS
             }
         ]
 
@@ -89,6 +95,7 @@ class NotifyEventsTest(base.NotifierTestCase):
 
         with db_api.transaction():
             wf_ex = db_api.get_workflow_execution(wf_ex.id)
+
             task_exs = wf_ex.task_executions
 
         self.assertEqual(states.SUCCESS, wf_ex.state)
@@ -113,8 +120,9 @@ class NotifyEventsTest(base.NotifierTestCase):
         self.assertIn((wf_ex.id, events.WORKFLOW_SUCCEEDED), EVENT_LOGS)
 
     def test_notify_all_implicit(self):
-        wf_def = """
+        wf_text = """
         version: '2.0'
+
         wf:
           tasks:
             t1:
@@ -125,7 +133,7 @@ class NotifyEventsTest(base.NotifierTestCase):
               action: std.noop
         """
 
-        wf_svc.create_workflows(wf_def)
+        wf_svc.create_workflows(wf_text)
 
         notify_options = [{'type': 'webhook'}]
         params = {'notify': notify_options}
@@ -160,8 +168,9 @@ class NotifyEventsTest(base.NotifierTestCase):
         self.assertIn((wf_ex.id, events.WORKFLOW_SUCCEEDED), EVENT_LOGS)
 
     def test_notify_order(self):
-        wf_def = """
+        wf_text = """
         version: '2.0'
+
         wf:
           tasks:
             t1:
@@ -172,11 +181,9 @@ class NotifyEventsTest(base.NotifierTestCase):
               action: std.noop
         """
 
-        wf_svc.create_workflows(wf_def)
+        wf_svc.create_workflows(wf_text)
 
-        notify_options = [
-            {'type': 'webhook'}
-        ]
+        notify_options = [{'type': 'webhook'}]
 
         params = {'notify': notify_options}
 
@@ -212,12 +219,10 @@ class NotifyEventsTest(base.NotifierTestCase):
         self.assertTrue(self.publishers['wbhk'].publish.called)
         self.assertListEqual(expected_order, EVENT_LOGS)
 
-    def test_notify_multiple(self):
-        self.assertFalse(self.publishers['wbhk'].publish.called)
-        self.assertFalse(self.publishers['noop'].publish.called)
-
-        wf_def = """
+    def test_notify_with_event_filter(self):
+        wf_text = """
         version: '2.0'
+
         wf:
           tasks:
             t1:
@@ -228,7 +233,63 @@ class NotifyEventsTest(base.NotifierTestCase):
               action: std.noop
         """
 
-        wf_svc.create_workflows(wf_def)
+        wf_svc.create_workflows(wf_text)
+
+        notify_options = [
+            {
+                'type': 'webhook',
+                'event_types': [
+                    events.WORKFLOW_LAUNCHED,
+                    events.WORKFLOW_SUCCEEDED
+                ]
+            }
+        ]
+
+        params = {'notify': notify_options}
+
+        wf_ex = self.engine.start_workflow('wf', '', **params)
+
+        self.await_workflow_success(wf_ex.id)
+
+        with db_api.transaction():
+            wf_ex = db_api.get_workflow_execution(wf_ex.id)
+            task_exs = wf_ex.task_executions
+
+        self.assertEqual(states.SUCCESS, wf_ex.state)
+        self.assertIsNone(wf_ex.state_info)
+        self.assertEqual(2, len(task_exs))
+
+        t1_ex = self._assert_single_item(task_exs, name='t1')
+        t2_ex = self._assert_single_item(task_exs, name='t2')
+
+        self.assertEqual(states.SUCCESS, t1_ex.state)
+        self.assertIsNone(t1_ex.state_info)
+        self.assertEqual(states.SUCCESS, t2_ex.state)
+        self.assertIsNone(t2_ex.state_info)
+
+        self.assertTrue(self.publishers['wbhk'].publish.called)
+        self.assertEqual(2, len(EVENT_LOGS))
+        self.assertIn((wf_ex.id, events.WORKFLOW_LAUNCHED), EVENT_LOGS)
+        self.assertIn((wf_ex.id, events.WORKFLOW_SUCCEEDED), EVENT_LOGS)
+
+    def test_notify_multiple(self):
+        self.assertFalse(self.publishers['wbhk'].publish.called)
+        self.assertFalse(self.publishers['noop'].publish.called)
+
+        wf_text = """
+        version: '2.0'
+
+        wf:
+          tasks:
+            t1:
+              action: std.noop
+              on-success:
+                - t2
+            t2:
+              action: std.noop
+        """
+
+        wf_svc.create_workflows(wf_text)
 
         notify_options = [
             {'type': 'webhook'},
@@ -280,8 +341,9 @@ class NotifyEventsTest(base.NotifierTestCase):
         self.assertFalse(self.publishers['wbhk'].publish.called)
         self.assertFalse(self.publishers['noop'].publish.called)
 
-        wf_def = """
+        wf_text = """
         version: '2.0'
+
         wf:
           tasks:
             t1:
@@ -292,7 +354,7 @@ class NotifyEventsTest(base.NotifierTestCase):
               action: std.noop
         """
 
-        wf_svc.create_workflows(wf_def)
+        wf_svc.create_workflows(wf_text)
 
         notify_options = [
             {'type': 'webhook'},
@@ -348,8 +410,9 @@ class NotifyEventsTest(base.NotifierTestCase):
         self.assertFalse(self.publishers['wbhk'].publish.called)
         self.assertFalse(self.publishers['noop'].publish.called)
 
-        wf_def = """
+        wf_text = """
         version: '2.0'
+
         wf:
           tasks:
             t1:
@@ -360,7 +423,7 @@ class NotifyEventsTest(base.NotifierTestCase):
               action: std.noop
         """
 
-        wf_svc.create_workflows(wf_def)
+        wf_svc.create_workflows(wf_text)
 
         cfg.CONF.set_default(
             'notify',
@@ -410,9 +473,11 @@ class NotifyEventsTest(base.NotifierTestCase):
         self.assertListEqual(expected_order, EVENT_LOGS)
 
     def test_workbook_notify(self):
-        wb_def = """
+        wb_text = """
         version: '2.0'
+
         name: wb
+
         workflows:
           wf1:
             tasks:
@@ -422,13 +487,14 @@ class NotifyEventsTest(base.NotifierTestCase):
                   - t2
               t2:
                 action: std.noop
+
           wf2:
             tasks:
               t1:
                 action: std.noop
         """
 
-        wb_svc.create_workbook_v2(wb_def)
+        wb_svc.create_workbook_v2(wb_text)
 
         notify_options = [{'type': 'webhook'}]
         params = {'notify': notify_options}
@@ -488,8 +554,9 @@ class NotifyEventsTest(base.NotifierTestCase):
         self.assertListEqual(expected_order, EVENT_LOGS)
 
     def test_notify_task_error(self):
-        wf_def = """
+        wf_text = """
         version: '2.0'
+
         wf:
           tasks:
             t1:
@@ -500,7 +567,7 @@ class NotifyEventsTest(base.NotifierTestCase):
               action: std.fail
         """
 
-        wf_svc.create_workflows(wf_def)
+        wf_svc.create_workflows(wf_text)
 
         notify_options = [{'type': 'webhook'}]
         params = {'notify': notify_options}
@@ -538,8 +605,9 @@ class NotifyEventsTest(base.NotifierTestCase):
         self.assertListEqual(expected_order, EVENT_LOGS)
 
     def test_notify_task_transition_fail(self):
-        wf_def = """
+        wf_text = """
         version: '2.0'
+
         wf:
           tasks:
             t1:
@@ -548,7 +616,7 @@ class NotifyEventsTest(base.NotifierTestCase):
                 - fail
         """
 
-        wf_svc.create_workflows(wf_def)
+        wf_svc.create_workflows(wf_text)
 
         notify_options = [{'type': 'webhook'}]
         params = {'notify': notify_options}
@@ -581,8 +649,9 @@ class NotifyEventsTest(base.NotifierTestCase):
         self.assertListEqual(expected_order, EVENT_LOGS)
 
     def test_notify_with_items_task(self):
-        wf_def = """
+        wf_text = """
         version: '2.0'
+
         wf:
           tasks:
             t1:
@@ -594,7 +663,7 @@ class NotifyEventsTest(base.NotifierTestCase):
               action: std.noop
         """
 
-        wf_svc.create_workflows(wf_def)
+        wf_svc.create_workflows(wf_text)
 
         notify_options = [{'type': 'webhook'}]
         params = {'notify': notify_options}
@@ -633,8 +702,9 @@ class NotifyEventsTest(base.NotifierTestCase):
         self.assertListEqual(expected_order, EVENT_LOGS)
 
     def test_notify_pause_resume(self):
-        wf_def = """
+        wf_text = """
         version: '2.0'
+
         wf:
           tasks:
             t1:
@@ -645,7 +715,7 @@ class NotifyEventsTest(base.NotifierTestCase):
               action: std.noop
         """
 
-        wf_svc.create_workflows(wf_def)
+        wf_svc.create_workflows(wf_text)
 
         notify_options = [{'type': 'webhook'}]
         params = {'notify': notify_options}
@@ -669,6 +739,7 @@ class NotifyEventsTest(base.NotifierTestCase):
 
         # Pause the workflow.
         self.engine.pause_workflow(wf_ex.id)
+
         self.await_workflow_paused(wf_ex.id)
 
         with db_api.transaction():
@@ -676,6 +747,7 @@ class NotifyEventsTest(base.NotifierTestCase):
             task_exs = wf_ex.task_executions
 
         t1_ex = self._assert_single_item(task_exs, name='t1')
+
         t1_act_exs = db_api.get_action_executions(task_execution_id=t1_ex.id)
 
         # Workflow is paused but the task is still running as expected.
@@ -760,8 +832,9 @@ class NotifyEventsTest(base.NotifierTestCase):
         self.assertListEqual(expected_order, EVENT_LOGS)
 
     def test_notify_pause_resume_task(self):
-        wf_def = """
+        wf_text = """
         version: '2.0'
+
         wf:
           tasks:
             t1:
@@ -772,7 +845,7 @@ class NotifyEventsTest(base.NotifierTestCase):
               action: std.noop
         """
 
-        wf_svc.create_workflows(wf_def)
+        wf_svc.create_workflows(wf_text)
 
         notify_options = [{'type': 'webhook'}]
         params = {'notify': notify_options}
@@ -796,6 +869,7 @@ class NotifyEventsTest(base.NotifierTestCase):
 
         # Pause the action execution of task 1.
         self.engine.on_action_update(t1_act_exs[0].id, states.PAUSED)
+
         self.await_workflow_paused(wf_ex.id)
 
         with db_api.transaction():
@@ -882,8 +956,9 @@ class NotifyEventsTest(base.NotifierTestCase):
         self.assertListEqual(expected_order, EVENT_LOGS)
 
     def test_notify_cancel(self):
-        wf_def = """
+        wf_text = """
         version: '2.0'
+
         wf:
           tasks:
             t1:
@@ -894,7 +969,7 @@ class NotifyEventsTest(base.NotifierTestCase):
               action: std.noop
         """
 
-        wf_svc.create_workflows(wf_def)
+        wf_svc.create_workflows(wf_text)
 
         notify_options = [{'type': 'webhook'}]
         params = {'notify': notify_options}
@@ -975,8 +1050,9 @@ class NotifyEventsTest(base.NotifierTestCase):
         self.assertListEqual(expected_order, EVENT_LOGS)
 
     def test_notify_cancel_task(self):
-        wf_def = """
+        wf_text = """
         version: '2.0'
+
         wf:
           tasks:
             t1:
@@ -987,7 +1063,7 @@ class NotifyEventsTest(base.NotifierTestCase):
               action: std.noop
         """
 
-        wf_svc.create_workflows(wf_def)
+        wf_svc.create_workflows(wf_text)
 
         notify_options = [{'type': 'webhook'}]
         params = {'notify': notify_options}
@@ -1035,3 +1111,256 @@ class NotifyEventsTest(base.NotifierTestCase):
 
         self.assertTrue(self.publishers['wbhk'].publish.called)
         self.assertListEqual(expected_order, EVENT_LOGS)
+
+    def test_notify_task_input_error(self):
+        wf_text = """---
+        version: '2.0'
+
+        wf:
+          tasks:
+            task1:
+              input:
+                url: <% $.ItWillBeError %>
+              action: std.http
+              on-error: task2
+            task2:
+              action: std.noop
+        """
+
+        wf_svc.create_workflows(wf_text)
+
+        notify_options = [{'type': 'webhook'}]
+        params = {'notify': notify_options}
+
+        wf_ex = self.engine.start_workflow('wf', '', **params)
+
+        self.await_workflow_error(wf_ex.id)
+
+        with db_api.transaction():
+            wf_ex = db_api.get_workflow_execution(wf_ex.id)
+            task_exs = wf_ex.task_executions
+
+        self.assertEqual(1, len(task_exs))
+
+        t1_ex = self._assert_single_item(task_exs, name='task1')
+        self.assertEqual(states.ERROR, t1_ex.state)
+
+        expected_order = [
+            (wf_ex.id, events.WORKFLOW_LAUNCHED),
+            (t1_ex.id, events.TASK_LAUNCHED),
+            (t1_ex.id, events.TASK_FAILED),
+            (wf_ex.id, events.WORKFLOW_FAILED)
+        ]
+
+        self.assertTrue(self.publishers['wbhk'].publish.called)
+        self.assertListEqual(expected_order, EVENT_LOGS)
+
+    @mock.patch('mistral.actions.std_actions.NoOpAction.run', mock.MagicMock(
+        side_effect=[Exception(), None, None]))
+    def test_notify_rerun_task(self):
+        wf_text = """
+        version: '2.0'
+
+        wf:
+          tasks:
+            t1:
+              action: std.noop
+              on-success:
+                - t2
+            t2:
+              action: std.noop
+        """
+
+        wf_svc.create_workflows(wf_text)
+
+        notify_options = [{'type': 'webhook'}]
+        params = {'notify': notify_options}
+
+        wf_ex = self.engine.start_workflow('wf', '', **params)
+
+        self.await_workflow_error(wf_ex.id)
+
+        with db_api.transaction():
+            wf_ex = db_api.get_workflow_execution(wf_ex.id)
+            task_exs = wf_ex.task_executions
+
+        t1_ex = self._assert_single_item(task_exs, name='t1')
+
+        self.assertEqual(states.ERROR, t1_ex.state)
+        self.assertEqual(1, len(task_exs))
+
+        # Rerun the workflow.
+        self.engine.rerun_workflow(t1_ex.id)
+
+        self.await_workflow_success(wf_ex.id)
+
+        with db_api.transaction():
+            wf_ex = db_api.get_workflow_execution(wf_ex.id)
+            task_exs = wf_ex.task_executions
+
+        t1_ex = self._assert_single_item(task_exs, name='t1')
+        t2_ex = self._assert_single_item(task_exs, name='t2')
+
+        self.assertEqual(states.SUCCESS, t1_ex.state)
+        self.assertEqual(states.SUCCESS, t2_ex.state)
+        self.assertEqual(2, len(task_exs))
+
+        expected_order = [
+            (wf_ex.id, events.WORKFLOW_LAUNCHED),
+            (t1_ex.id, events.TASK_LAUNCHED),
+            (t1_ex.id, events.TASK_FAILED),
+            (wf_ex.id, events.WORKFLOW_FAILED),
+            # rerun
+            (wf_ex.id, events.WORKFLOW_RERUN),
+            (t1_ex.id, events.TASK_RERUN),
+            (t1_ex.id, events.TASK_SUCCEEDED),
+            (t2_ex.id, events.TASK_LAUNCHED),
+            (t2_ex.id, events.TASK_SUCCEEDED),
+            (wf_ex.id, events.WORKFLOW_SUCCEEDED),
+        ]
+
+        self.assertTrue(self.publishers['wbhk'].publish.called)
+        self.assertListEqual(expected_order, EVENT_LOGS)
+
+    @mock.patch('mistral.actions.std_actions.NoOpAction.run', mock.MagicMock(
+        side_effect=[Exception(), None, None, None]))
+    def test_notify_rerun_nested_workflow(self):
+        wf_text = """
+        version: '2.0'
+
+        wf_1:
+          tasks:
+            wf_1_t1:
+              workflow: wf_2
+              on-success:
+                - wf_1_t2
+
+            wf_1_t2:
+              action: std.noop
+
+        wf_2:
+          tasks:
+            wf_2_t1:
+              action: std.noop
+              on-success:
+                - wf_2_t2
+
+            wf_2_t2:
+              action: std.noop
+        """
+
+        wf_svc.create_workflows(wf_text)
+
+        notify_options = [{'type': 'webhook'}]
+        params = {'notify': notify_options}
+
+        wf_1_ex = self.engine.start_workflow('wf_1', '', **params)
+
+        self.await_workflow_error(wf_1_ex.id)
+
+        with db_api.transaction():
+            wf_exs = db_api.get_workflow_executions()
+
+            self._assert_single_item(
+                wf_exs,
+                name='wf_1',
+                state=states.ERROR
+            )
+            self._assert_single_item(
+                wf_exs,
+                name='wf_2',
+                state=states.ERROR
+            )
+
+            task_exs = db_api.get_task_executions()
+
+            self._assert_single_item(
+                task_exs,
+                name='wf_1_t1',
+                state=states.ERROR
+            )
+
+            wf_2_t1 = self._assert_single_item(
+                task_exs,
+                name='wf_2_t1',
+                state=states.ERROR
+            )
+
+        self.assertEqual(2, len(task_exs))
+        self.assertEqual(2, len(wf_exs))
+
+        # Rerun the nested workflow.
+        self.engine.rerun_workflow(wf_2_t1.id)
+
+        self.await_workflow_success(wf_1_ex.id)
+
+        with db_api.transaction():
+            wf_exs = db_api.get_workflow_executions()
+
+            wf_1_ex = self._assert_single_item(
+                wf_exs,
+                name='wf_1',
+                state=states.SUCCESS
+            )
+            wf_2_ex = self._assert_single_item(
+                wf_exs,
+                name='wf_2',
+                state=states.SUCCESS
+            )
+
+            task_wf_1_exs = wf_1_ex.task_executions
+
+            wf_1_t1 = self._assert_single_item(
+                task_wf_1_exs,
+                name='wf_1_t1',
+                state=states.SUCCESS
+            )
+            wf_1_t2 = self._assert_single_item(
+                task_wf_1_exs,
+                name='wf_1_t2',
+                state=states.SUCCESS
+            )
+
+            task_wf_2_exs = wf_2_ex.task_executions
+
+            wf_2_t1 = self._assert_single_item(
+                task_wf_2_exs,
+                name='wf_2_t1',
+                state=states.SUCCESS
+            )
+            wf_2_t2 = self._assert_single_item(
+                task_wf_2_exs,
+                name='wf_2_t2',
+                state=states.SUCCESS
+            )
+
+            self.assertEqual(2, len(task_wf_1_exs))
+            self.assertEqual(2, len(task_wf_2_exs))
+            self.assertEqual(2, len(wf_exs))
+
+            expected_order = [
+                (wf_1_ex.id, events.WORKFLOW_LAUNCHED),
+                (wf_1_t1.id, events.TASK_LAUNCHED),
+                (wf_2_ex.id, events.WORKFLOW_LAUNCHED),
+                (wf_2_t1.id, events.TASK_LAUNCHED),
+                (wf_2_t1.id, events.TASK_FAILED),
+                (wf_2_ex.id, events.WORKFLOW_FAILED),
+                (wf_1_t1.id, events.TASK_FAILED),
+                (wf_1_ex.id, events.WORKFLOW_FAILED),
+                # rerun
+                (wf_2_ex.id, events.WORKFLOW_RERUN),
+                (wf_1_ex.id, events.WORKFLOW_RERUN),
+                (wf_1_t1.id, events.TASK_RERUN),
+                (wf_2_t1.id, events.TASK_RERUN),
+                (wf_2_t1.id, events.TASK_SUCCEEDED),
+                (wf_2_t2.id, events.TASK_LAUNCHED),
+                (wf_2_t2.id, events.TASK_SUCCEEDED),
+                (wf_2_ex.id, events.WORKFLOW_SUCCEEDED),
+                (wf_1_t1.id, events.TASK_SUCCEEDED),
+                (wf_1_t2.id, events.TASK_LAUNCHED),
+                (wf_1_t2.id, events.TASK_SUCCEEDED),
+                (wf_1_ex.id, events.WORKFLOW_SUCCEEDED),
+            ]
+
+            self.assertTrue(self.publishers['wbhk'].publish.called)
+            self.assertListEqual(expected_order, EVENT_LOGS)
